import logging
from typing import Tuple
from pathlib import Path

import biocypher
import yaml
import pandas as pd
import pandera as pa

from . import base
Node = base.Node
Edge = base.Edge
Transformer = base.Transformer
Adapter = base.Adapter
All = base.All

from . import types
from . import transformer
from . import tabular
from . import serialize
from . import congregate
from . import merge
from . import fuse
from . import fusion
from . import exceptions
from . import owl_to_biocypher
from . import biocypher_to_owl
from . import make_value
from . import make_labels

logger = logging.getLogger("ontoweaver")

__all__ = ['Node', 'Edge', 'Transformer', 'Adapter', 'All', 'tabular', 'types', 'transformer', 'serialize', 'congregate',
           'merge', 'fuse', 'fusion', 'exceptions', 'logger', 'owl_to_biocypher', 'biocypher_to_owl', 'make_value', "make_labels"]

def read_file(filename, **kwargs):
    """Read a file with Pandas, using its extension to guess its format.

    If no additional arguments are passed, it will call the
    Pandas `read_*` function with `filter_na = False`, which makes empty cell
    values to be loaded as empty strings instead of NaN values.

    Args:
        filename: The name of the data file the user wants to map.
        separator (str, optional): The separator used in the data file. Defaults to None.
        kwargs: A dictionary of arguments to pass to pandas.read_* functions.

    Raises:
        exception.FeatureError: if the extension is unknown.

    Returns:
        A Pandas DataFrame.
    """

    # We probably don't want NaN as a default,
    # since they tend to end up in a label.
    kwargs.update({'na_filter': True,
                'engine': 'python'}) #'c' engine does not support regex separators (separators > 1 char and different
                                  # from '\s+' are interpreted as regex) which results in an error.

    read_funcs = {
        '.csv'    : pd.read_csv,
        '.tsv'    : pd.read_csv,
        '.txt'    : pd.read_csv,

        '.xls'    : pd.read_excel,
        '.xlsx'   : pd.read_excel,
        '.xlsm'   : pd.read_excel,
        '.xlsb'   : pd.read_excel,
        '.odf'    : pd.read_excel,
        '.ods'    : pd.read_excel,
        '.odt'    : pd.read_excel,

        '.json'   : pd.read_json,
        '.html'   : pd.read_html,
        '.xml'    : pd.read_xml,
        '.hdf'    : pd.read_hdf,
        '.feather': pd.read_feather,
        '.parquet': pd.read_parquet,
        '.pickle' : pd.read_pickle,
        '.orc'    : pd.read_orc,
        '.sas'    : pd.read_sas,
        '.spss'   : pd.read_spss,
        '.stata'  : pd.read_stata,
    }
    filepath = Path(filename)
    ext = filepath.suffix

    if ext not in read_funcs:
        msg = f"File format '{ext}' of file '{filename}' is not supported (I can only read one of: {' ,'.join(read_funcs.keys())})"
        logger.error(msg)
        raise exceptions.FeatureError(msg)

    return read_funcs[ext](filename, **kwargs)


def extract_reconciliate_write(biocypher_config_path, schema_path, filename_to_mapping = None, dataframe_to_mapping = None, parallel_mapping = 0, separator = None, affix = "none", affix_separator = ":", validate_output = False, raise_errors = True, **kwargs):
    """Calls several mappings, each on the related Pandas-readable tabular data file,
       then reconciliate duplicated nodes and edges (on nodes' IDs, merging properties in lists),
       then export everything with BioCypher.
       Returns the path to the resulting import file.

       Args:
           biocypher_config_path: the BioCypher configuration file.
           schema_path: the assembling schema file.
           filename_to_mapping: a dictionary mapping data file path to the OntoWeaver mapping yaml file to extract them.
           dataframe_to_mapping (tuple): Tuple containing pairs of loaded Pandas DataFrames and their corresponding loaded YAML mappings.
           parallel_mapping (int): Number of workers to use in parallel mapping. Defaults to 0 for sequential processing.
           separator (str, optional): The separator to use for combining values in reconciliation. Defaults to None.
           affix (str, optional): The affix to use for type inclusion. Defaults to "none".
           affix_separator: The character(s) separating the label from its type affix. Defaults to ":".
           validate_output: Whether to validate the output of the transformers. Defaults to False.
           raise_errors: Whether to raise errors encountered during the mapping, and stop the mapping process. Defaults to True.
           kwargs: A dictionary of arguments to pass to pandas.read_* functions.

       Returns:
           The path to the import file.
   """
    nodes = []
    edges = []

    if filename_to_mapping:

        assert(type(filename_to_mapping) == dict) # data_file => mapping_file

        for data_file, mapping_file in filename_to_mapping.items():
            table = read_file(data_file, **kwargs)

            with open(mapping_file) as fd:
                mapping = yaml.full_load(fd)

            adapter = tabular.extract_table(
                table,
                mapping,
                parallel_mapping=parallel_mapping,
                affix=affix,
                separator=affix_separator,
                validate_output=validate_output,
                raise_errors = raise_errors,
            )

            nodes += adapter.nodes
            edges += adapter.edges

    if dataframe_to_mapping:

        for dataframe, loaded_mapping in dataframe_to_mapping:
            adapter = tabular.extract_table(
                dataframe,
                loaded_mapping,
                parallel_mapping=parallel_mapping,
                affix=affix,
                separator=affix_separator,
                validate_output=validate_output,
                raise_errors=raise_errors,
            )

            nodes += adapter.nodes
            edges += adapter.edges

    fnodes, fedges = fusion.reconciliate(nodes, edges, separator = separator)

    bc = biocypher.BioCypher(    # fixme change constructor to take contents of paths instead of reading path.
        biocypher_config_path = biocypher_config_path,
        schema_config_path = schema_path
    )

    if fnodes:
        bc.write_nodes(fnodes)
    if fedges:
        bc.write_edges(fedges)
    import_file = bc.write_import_call()

    return import_file


def extract(filename_to_mapping = None, dataframe_to_mapping = None, parallel_mapping = 0, affix="none", affix_separator=":", validate_output = False, raise_errors = True, **kwargs) -> Tuple[list[Tuple], list[Tuple]]:
    """
    Extracts nodes and edges from tabular data files based on provided mappings.

    Args:
        filename_to_mapping (dict): A dictionary mapping data file path to the OntoWeaver mapping yaml file to extract them.
        dataframe_to_mapping (tuple): Tuple containing pairs of loaded Pandas DataFrames and their corresponding loaded YAML mappings.
        parallel_mapping (int): Number of workers to use in parallel mapping. Defaults to 0 for sequential processing.
        affix (str, optional): The affix to use for type inclusion. Defaults to "none".
        affix_separator: The character(s) separating the label from its type affix. Defaults to ":".
        validate_output: Whether to validate the output of the transformers. Defaults to False.
        raise_errors: Whether to raise errors encountered during the mapping, and stop the mapping process. Defaults to True.
        kwargs: A dictionary of arguments to pass to pandas.read_* functions.

    Returns:
        tuple: Two lists of tuples containing nodes and edges.
    """

    nodes = []
    edges = []

    if filename_to_mapping:

        assert(type(filename_to_mapping) == dict) # data_file => mapping_file

        for data_file, mapping_file in filename_to_mapping.items():
            table = read_file(data_file, **kwargs)

            with open(mapping_file) as fd:
                mapping = yaml.full_load(fd)

            adapter = tabular.extract_table(
                table,
                mapping,
                parallel_mapping=parallel_mapping,
                affix=affix,
                separator=affix_separator,
                validate_output=validate_output,
                raise_errors = raise_errors,
            )

            nodes += adapter.nodes
            edges += adapter.edges

    if dataframe_to_mapping:

        for dataframe, loaded_mapping in dataframe_to_mapping:

            adapter = tabular.extract_table(
                dataframe,
                loaded_mapping,
                parallel_mapping=parallel_mapping,
                affix=affix,
                separator=affix_separator,
                validate_output=validate_output,
                raise_errors=raise_errors,
            )

            nodes += adapter.nodes
            edges += adapter.edges

    return nodes, edges


def reconciliate_write(nodes: list[Tuple], edges: list[Tuple], biocypher_config_path: str, schema_path: str, separator: str = None, raise_errors = True) -> str:
    """
    Reconciliates duplicated nodes and edges, then writes them using BioCypher.

    Args:
        nodes (list): A list of nodes to be reconciliated and written.
        edges (list): A list of edges to be reconciliated and written.
        biocypher_config_path (str): the BioCypher configuration file.
        schema_path (str): the assembling schema file
        separator (str, optional): The separator to use for combining values in reconciliation. Defaults to None.

        FIXME: The raise_errors parameter is currently not used downstream because the fusion classes are to be refactored with error management.
        raise_errors: Whether to raise errors encountered during the mapping, and stop the mapping process. Defaults to True.

    Returns:
        str: The path to the import file.
    """

    fnodes, fedges = fusion.reconciliate(nodes, edges, separator = separator)

    bc = biocypher.BioCypher(
        biocypher_config_path = biocypher_config_path,
        schema_config_path = schema_path
    )

    if fnodes:
        bc.write_nodes(fnodes)
    if fedges:
        bc.write_edges(fedges)
    #bc.summary()
    import_file = bc.write_import_call()

    return import_file


def validate_input_data(filename_to_mapping: dict, raise_errors = True, **kwargs) -> bool:
    """
    Validates the data files based on provided rules in configuration.

    Args:
        filename_to_mapping (dict): a dictionary mapping data file path to the OntoWeaver mapping yaml file.
        raise_errors: Whether to raise errors encountered during the mapping, and stop the mapping process. Defaults to True.
        kwargs: A dictionary of arguments to pass to pandas.read_* functions.

    Returns:
        bool: True if the data is valid, False otherwise.
    """

    assert(type(filename_to_mapping) == dict) # data_file => mapping_file

    for data_file, mapping_file in filename_to_mapping.items():
        table = read_file(data_file, **kwargs)

        with open(mapping_file) as fd:
            yaml_mapping = yaml.full_load(fd)

        validator = tabular.YamlParser(yaml_mapping, types, raise_errors=raise_errors)._get_input_validation_rules()

        try:
            validator(table)
            return True
        except pa.errors.SchemaErrors as exc:
            logger.error(f"Validation failed for {exc.failure_cases}.")
            return False
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            return False


def validate_input_data_loaded(dataframe, loaded_mapping, raise_errors = True) -> bool:
    """
    Validates the data files based on provided rules in configuration.

    Args:
         dataframe: The loaded pandas DataFrame to validate.
         loaded_mapping: The mapping object to use for validation.
         raise_errors: Whether to raise errors encountered during the mapping, and stop the mapping process. Defaults to True.

    Returns:
        bool: True if the data is valid, False otherwise.
    """

    validator = tabular.YamlParser(loaded_mapping, types, raise_errors=raise_errors)._get_input_validation_rules()

    try:
        validator(dataframe)
        return True
    except pa.errors.SchemaErrors as exc:
        logger.error(f"Validation failed for {exc.failure_cases}.")
        return False
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return False




import logging
import abc
import re

from . import errormanager, exceptions

logger = logging.getLogger("ontoweaver")

class ReturnCreate:
    """Class for returning the extracted cell value of the transformer, along with the edge type, target node type,
    and properties of the target node type.
    """

    def __init__(self, extracted_cell_value = None, edge_type = None, target_node_type = None, properties_of = None, final_type = None):
        """
        Initializes the ReturnCreate object.

        Args:
            extracted_cell_value: The extracted cell value of the transformer. Can serve as ID of the node, ID of the edge, or value of property.
            edge_type: The type of the edge.
            target_node_type: The type of the target node.
            properties_of: The properties of the returned target node type.
            final_type: The final type of the target node.
        """

        self.extracted_cell_value = extracted_cell_value # Can serve as ID of the node, ID of the edge, or value of property.
        self.edge_type = edge_type
        self.target_node_type = target_node_type
        self.target_element_properties = properties_of
        self.final_type = final_type

class LabelMaker(errormanager.ErrorManager, metaclass=abc.ABCMeta):
    """Interface for defining the correct labels (types) for each extracted value.
    The class is responsible for validating the extracted cell value and returning the appropriate label.
    """

    def __init__(self, raise_errors: bool = True):
        """
        Initializes the LabelMaker object.

        Args:
            raise_errors (bool): if True, will raise an exception when an error is encountered, else, will log the error and try to proceed.
        """
        super().__init__(raise_errors)

    @abc.abstractmethod
    def __call__(self, validate, returned_value, multi_type_dict, branching_properties = None, row = None):
        """

        Args:
            validate: The validation function which uses the Pandera schema to validate the extracted value.
            returned_value: The extracted value of the cell.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            branching_properties: In case of branching on cell values, the dictionary holding the properties for each branch.
            row: The row of the dataframe that is being processed. Used in case of branching on column values.

        Returns:
            ReturnCreate: An object containing the defined values.
        """
        return NotImplementedError("The call method must be implemented in a subclass.")


class SimpleLabelMaker(LabelMaker):
    """
    The class is used when the transformer does not have any kind of type branching logic.
    """

    def __init__(self, raise_errors: bool = True):
        super().__init__(raise_errors)

    def __call__(self, validate, returned_value, multi_type_dict, branching_properties = None, row = None):

        res = str(returned_value)
        if validate(res):
            if multi_type_dict:
                if "None" in multi_type_dict.keys():
                    # No branching needed. The transformer is not a branching transformer.
                    return ReturnCreate(res, multi_type_dict["None"]["via_relation"], multi_type_dict["None"]["to_object"], None, multi_type_dict["None"]["final_type"])
            else:
                # No multi-type dictionary. The transformer returns only the extracted value of the cell. Used for properties.
                return ReturnCreate(res)
        else:
            # Validation failed, return empty object with None values.
            return ReturnCreate()

class MultiTypeLabelMaker(LabelMaker):
    """
    The class is used when the transformer has type branching logic based on the value that will become the ID of the element.
    """
    def __init__(self,raise_errors: bool = True):
        super().__init__(raise_errors)

    def __call__(self, validate, returned_value, multi_type_dict = None, branching_properties = None, row = None):
        res = str(returned_value)
        if validate(res):
            if multi_type_dict:
                for key, types in multi_type_dict.items():
                    # Branching is performed on the regex patterns.
                    if re.search(key, res):
                        if branching_properties:
                            properties_of = branching_properties.get(types["to_object"].__name__, {})
                        else:
                            properties_of = {}
                        return ReturnCreate(res, types["via_relation"], types["to_object"], properties_of, types["final_type"])
                    else:
                        logger.info(f"          Branching key `{key}` does not match extracted value `{res}`.")
                        continue
            else:
                # No multi-type dictionary. The transformer returns only the extracted value of the cell. Used for properties.
                return ReturnCreate(res)
        else:
            # Validation failed, return empty object with None values.
            return ReturnCreate()

class MultiTypeOnColumnLabelMaker(LabelMaker):
    """
    The class is used when the transformer has type branching logic based on the value of a column in the dataframe,
    not the value that will become the ID of the element.
    """
    def __init__(self, raise_errors: bool = True, match_type_from_column: str = None):
        self.match_type_from_column = match_type_from_column
        super().__init__(raise_errors)

    def __call__(self, validate, returned_value, multi_type_dict = None, branching_properties = None, row = None):
        res = str(returned_value)
        if validate(res):
            if multi_type_dict:
                for key, types in multi_type_dict.items():
                    # Branching is performed on the regex patterns.
                    if re.search(key, row[self.match_type_from_column]):
                        if branching_properties:
                            properties_of = branching_properties.get(types["to_object"].__name__, {})
                        else:
                            properties_of = {}
                        return ReturnCreate(res, types["via_relation"], types["to_object"], properties_of, types["final_type"])
                    else:
                        logger.info(f"          Branching key `{key}` does not match extracted value `{row[self.match_type_from_column]}`.")
                        continue
            else:
                # No multi-type dictionary. The transformer returns only the extracted value of the cell. Used for properties.
                return ReturnCreate(res)
        else:
            return ReturnCreate()






#!/usr/bin/env python3

import os
import sys
import logging
import pathlib
import platform
import xdg_base_dirs as xdg
import importlib
import networkx

error_codes = {
    "ParsingError"    :  65, # "data format"
    "RunError"        :  70, # "internal"
    "DataValidationError": 76,  # "protocol"
    "ConfigError"     :  78, # "bad config"
    "CannotAccessFile": 126, # "no perm"
    "FileError"       : 127, # "not found"
    "SubprocessError" : 128, # "bad exit"
    "NetworkXError"   : 129, # probably "type not in the digraph"
    "OntoWeaverError" : 254,
    "Exception"       : 255,
}

def check_file(filename):
    """Exit if the given filename does not exists or is not readable."""
    if not os.path.isfile(filename):
        logging.error(f"File `{filename}` not found.")
        sys.exit(error_codes["FileError"])

    if not os.access(filename, os.R_OK):
        logging.error(f"Cannot access file `{filename}`.")
        sys.exit(error_codes["CannotAccessFile"])


def config_directories(appname = "ontoweave"):
    """Yield standard configuration directories (as defined by XDG under MocOS/Unix)."""
    os = platform.system()
    logging.debug(f"Detected OS: {os}")

    #NOTE: All matched config files will be parsed and applied in the given order.
    if os == "Windows":
        if "APPDATA" in os.environ:
            yield pathlib.Path(os.environ["APPDATA"])
        yield pathlib.Path("~")/pathlib.Path("AppData")/pathlib.Path("Roaming")/pathlib.Path(appname)

    elif os == "Java" or os == "":
        logging.warning(f"I don't know where to search for configuration files on platform `{os}`, I'll only search in current directory")

    else: # Probably an Unix flavor (Darwin, Linux, Solaris, IRIX, etc.)
        # XDG will return the default defined in the specification
        # if the env variables are not set, so hardcoded defaults
        # like "~/.config/" should not be necessary.
        for p in xdg.xdg_config_dirs():
            yield p/appname
        yield xdg.xdg_config_home()/appname

    yield pathlib.Path("../ontoweaver")


def config_paths(appname = "ontoweave"):
    """Yield any path named <appname>.yaml in standard configuration directories."""
    dirs = config_directories()
    for p in dirs:
        yield str(p / (appname+".yaml"))


def import_from_path(file_path):
    """Import the given Python file path as a module."""
    # See https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    module_name = pathlib.Path(file_path).stem
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def main():
    import jsonargparse
    import argparse
    import subprocess
    import inspect

    appname = os.path.splitext(os.path.basename(sys.argv[0]))[0]

    logging.basicConfig()
    logger = logging.getLogger(appname)

    config_files = list(config_paths(appname))

    do = jsonargparse.ArgumentParser(
        description = "A command line tool to run OntoWeaver mapping adapters on a set of tabular data, and call the created BioCypher export scripts.",
        epilog = f"Example usage:\n  {appname} table.csv:mapping.yaml\n  {appname} --biocypher-config biocypher_config.yaml --biocypher-schema schema.yaml table1.csv:mapping1.yaml table2.tsv:mapping1.yaml table3.parquet:mapping2.yaml --import-script-run",
        default_config_files = config_files,
        env_prefix="ONTOWEAVE",
        default_env = True, # By default jsonargparse does not check environment, so one enables it explicitly.
        formatter_class = argparse.RawTextHelpFormatter,
        logger = logger, # FIXME jsonargparse seems to override any later config of logging, this seems like a bug.
    )

    do.add_argument("mapping", metavar="FILE:MAPPING", nargs="+",
        help=f"Run the given YAML MAPPING to extract data from the tabular FILE (usually a CSV). Several mappings can be passed to {appname}. You may also use the same mapping on different data files. If set to `STDIN`, will read the list of mappings from standard input.")

    do.add_argument("-c", "--config", metavar="FILE", action=jsonargparse.ActionConfigFile,
        help=f"The {appname} configuration file, which can host the same arguments than the command line tool. [default: {' or '.join(config_files)}]")

    do.add_argument("-C", "--biocypher-config", metavar="FILE", default="biocypher_config.yaml",
        help="The BioCypher config file (the one managing ontologies and the output backend). [default: %(default)s]")

    do.add_argument("-s", "--biocypher-schema", metavar="FILE", default="schema.yaml",
        help="The BioCypher schema file (the one managing node and edge types). [default: %(default)s]")

    do.add_argument("-p", "--parallel", metavar="NB_CORES", default="0",
        help=f"Number of processor cores to use when processing with multi-threading. `0` means a sequential processing (no parallelization, the default). Use 'auto' to let {appname} do its best to use a good number. [default: %(default)s]")

    do.add_argument("-i", "--import-script-run", action="store_true",
        help=f"If passed {appname} will call the import scripts created by Biocypher for you.")

    do.add_argument("-r", "--register", metavar="PYTHON_MODULE", nargs="*", default=[],
        help="Register all transformers available in the given module.")

    do.add_argument("-S", "--prop-sep", metavar="CHARACTER", default = ";",
        help="The character used to separate property values fused in the same property at the reconciliation step. [default: %(default)s]")

    do.add_argument("-a", "--type-affix", choices=["suffix","prefix","none"], default="none",
        help="Where to add the type string to the ID label. [default: %(default)s]")

    do.add_argument("-A", "--type-affix-sep", metavar="CHARACTER", default=":",
        help="Character used to separate the label from the type affix. [default: %(default)s]")

    do.add_argument("-E", "--pass-errors", action="store_true",
        help=f"When an error occurs, log is, and then try to continue processing. If not passed, the default behavior is to raise errors immediatly and stop execution.")

    do.add_argument("-l", "--log-level", default="WARNING",
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        help="Configure the log level. [default: %(default)s]")

    do.add_argument("-v", "--validate-only", action="store_true",
                    help="Only validate the given input data, do not apply the mapping.")

    do.add_argument("-V", "--validate-output", action="store_true",
                    help="Validate the output data against the mapping rules.")

    do.add_argument("-Ds", "--database-sep", metavar="CHARACTER",
        help="Character used to separate values in the database.")

    do.add_argument("-D", "--debug", action="store_true",
        help=f"Run in debug mode. implies `--log-level DEBUG`, disables `--pass-errors`. NOTE: this will disable explicit return codes and show the call stack.")

    asked = do.parse_args()

    if asked.debug:
        if asked.log_level != "DEBUG":
            logger.setLevel("DEBUG")
            logger.warning(f"You asked for --debug but set --log-level={asked.log_level}, I will ignore that and set --log-level=DEBUG")
        if asked.pass_errors:
            logger.warning(f"You asked for --debug but passed --pass-errors, I will ignore that.")
        asked.log_level = "DEBUG"
        asked.pass_errors = False

    logger.setLevel(asked.log_level)
    logging.getLogger("ontoweaver").setLevel(asked.log_level)

    logger.info("OntoWeave parameters:")

    logger.info(f"    config files: {config_files}")
    logger.info(f"    config: `{asked.biocypher_config}`")
    logger.info(f"    schema: `{asked.biocypher_schema}`")
    logger.info(f"    prop-sep: `{asked.prop_sep}`")
    logger.info(f"    type-affix: `{asked.type_affix}`")
    logger.info(f"    type-affix-sep: `{asked.type_affix_sep}`")
    logger.info(f"    import-script-run: `{asked.import_script_run}`")
    logger.info(f"    debug: `{asked.debug}`")
    logger.info(f"    log-level: `{asked.log_level}`")
    logger.info(f"    pass-errors: `{asked.pass_errors}`")
    logger.info(f"    validate-only: `{asked.validate_only}`")
    logger.info(f"    validate-output: `{asked.validate_output}`")

    logger.info(f"    asked mappings: `{asked.mapping}`")
    asked_mapping = []

    if asked.mapping == ["STDIN"]:
        while True:
            try:
                item = sys.stdin.readline()
            except UnicodeDecodeError:
                continue
            except KeyboardInterrupt:
                break
            if not item:
                break
            asked_mapping.append(item)
    else:
        asked_mapping = asked.mapping

    logger.info(f"    parsed mappings:")
    mappings = {}
    for data_map in asked_mapping:
        if ":" not in data_map:
            msg = f"Cannot parse the DATA:MAPPING `{data_map}`, I cannot find the colon character."
            logger.error(msg)
            sys.exit(error_codes["ConfigError"])
        data,map = data_map.split(":")
        mappings[data] = map
        logger.info(f"    `{data}` => `{map}`")

    if asked.parallel == "auto":
        parallel = min(32, (os.process_cpu_count() or 1) + 4)
    else:
        parallel = int(asked.parallel)
    logger.info(f"    parallel: `{asked.parallel}`")

    # Late import to avoid useless Biocypher's logs when asking for --help.
    import ontoweaver
    from biocypher._logger import get_logger as biocypher_logger
    biocypher_logger("biocypher").setLevel(asked.log_level)

    # Validate the input data if asked.
    if asked.validate_only:
        logger.info(f"Validating input data frame...")
        if ontoweaver.validate_input_data(filename_to_mapping=mappings, sep=asked.database_sep):
            logger.info(f"  Input data is valid according to provided rules.")
            sys.exit(0)
        else:
            logger.error(f"  Input data is invalid according to provided rules.")
            sys.exit(error_codes["DataValidationError"])

    # Register all transformers existing in the given modules.
    for mpath in asked.register:
        check_file(mpath)
        logger.info(f"Look for transformers in `{mpath}`")
        mod = import_from_path(mpath)
        for name,cls in mod.__dict__.items():
            if inspect.isclass(cls):
                logger.debug(f"{cls}")
                if issubclass(cls, ontoweaver.base.Transformer):
                    logger.info(f"    Register transformer: `{cls}`")
                    ontoweaver.transformer.register(cls)

    # Double check file inputs and exit on according errors.
    check_file(asked.biocypher_config)
    check_file(asked.biocypher_schema)
    for file_map in asked.mapping:
        data_file, map_file = file_map.split(":")
        check_file(data_file)
        check_file(map_file)

    if asked.validate_output:
        validate_output = True
    else:
        validate_output = False

    logger.info(f"Running OntoWeaver...")
    if asked.debug:
        import_file = ontoweaver.extract_reconciliate_write(
            asked.biocypher_config,
            asked.biocypher_schema,
            mappings,
            parallel_mapping=parallel,
            separator=asked.prop_sep,
            affix=asked.type_affix,
            affix_separator = asked.type_affix_sep,
            validate_output = validate_output,
            raise_errors = not asked.pass_errors)
    else:
        try:
            import_file = ontoweaver.extract_reconciliate_write(
                asked.biocypher_config,
                asked.biocypher_schema,
                mappings,
                parallel_mapping=parallel,
                separator=asked.prop_sep,
                affix=asked.type_affix,
                affix_separator = asked.type_affix_sep,
                validate_output = validate_output,
                raise_errors = not asked.pass_errors)
        # Manage exceptions wih specific error codes:
        except ontoweaver.exceptions.ConfigError as e:
            logger.error(f"ERROR in configuration: "+str(e))
            sys.exit(error_codes["ConfigError"])
        except ontoweaver.exceptions.RunError as e:
            logger.error(f"ERROR in content: "+str(e))
            sys.exit(error_codes["RunError"])
        except ontoweaver.exceptions.ParsingError as e:
            logger.error(f"ERROR during parsing of the YAML mapping: "+str(e))
            sys.exit(error_codes["ParsingError"])
        except ontoweaver.exceptions.DataValidationError as e:
            logger.error(f"ERROR during data validation: "+str(e))
            sys.exit(error_codes["DataValidationError"])
        except ontoweaver.exceptions.OntoWeaverError as e:
            logger.error(f"ERROR: "+str(e))
            sys.exit(error_codes["OntoWeaverError"])
        except networkx.exception.NetworkXError as e:
            logger.error(f"ERROR: "+str(e))
            logger.error("Double check that you use the rdfs:label for this type in your schema, and not the IRI anchor, or look for any typo.")
            sys.exit(error_codes["NetworkXError"])
        except Exception as e:
            logger.error(f"UNKNOWN ERROR: "+str(e))
            sys.exit(error_codes["Exception"])

    # Output import file on stdout, in case the user would want to capture it.
    print(import_file)
    check_file(import_file)

    if asked.import_script_run:
        shell = os.environ["SHELL"]
        logger.info(f"Run import scripts with {shell}...")
        if asked.debug:
            subprocess.run([shell, import_file])

        else:
            try:
                subprocess.run([shell, import_file])
            except Exception as e:
                logger.error(e)
                sys.exit(error_codes["SubprocessError"])


    logger.info("Done")

if __name__ == "__main__":
    main()




class OntoWeaverError(Exception):
    pass

class FeatureError(OntoWeaverError):
    pass

class ConfigError(OntoWeaverError):
    pass

class RunError(OntoWeaverError):
    pass

class DeclarationError(RunError):
    pass

class TransformerError(RunError):
    pass

class TransformerInterfaceError(TransformerError):
    pass

class TransformerDataError(TransformerError):
    pass

class DataValidationError(TransformerError):
    pass

class ParsingError(OntoWeaverError):
    pass

class ParsingPropertiesError(ParsingError):
    pass

class MissingFieldError(ParsingPropertiesError):
    pass

class ParsingDeclarationsError(ParsingError):
    pass

class CardinalityError(ParsingDeclarationsError):
    pass

class MissingDataError(ParsingDeclarationsError):
    pass


class InterfaceInheritanceError(TransformerError):
    pass

class TransformerConfigError(TransformerError):
    pass

class TransformerInputError(TransformerError):
    pass




import math
import yaml
import logging
import threading

import types as pytypes
import pandas as pd
import pandera as pa

from itertools import chain
from concurrent.futures import ThreadPoolExecutor
from typing import Optional
from collections.abc import Iterable
from enum import Enum, EnumMeta

from . import errormanager
from . import base
from . import types
from . import transformer
from . import exceptions
from . import validate
from . import make_labels

logger = logging.getLogger("ontoweaver")

class MetaEnum(EnumMeta):
    """
    Metaclass for Enum to allow checking if an item is in the Enum.
    """

    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        return True


class Enumerable(Enum, metaclass=MetaEnum):
    """
    Base class for Enums with MetaEnum metaclass.
    """
    pass


class TypeAffixes(str, Enumerable):
    """
    Enum for type affixes used in ID creation.
    """
    suffix = "suffix"
    prefix = "prefix"
    none = "none"


class PandasAdapter(base.Adapter):
    """Interface for extracting data from a Pandas DataFrame with a simple mapping configuration based on declared types.

    The general idea is that each row of the table is mapped to a source node,
    and some column values are mapped to an edge leading to another node.
    Some other columns may also be mapped to properties of either a node or an edge.

    The class expect a configuration formed by three objects:
        - the type of the source node mapped for each row.
        - a dictionary mapping each column name to the type of the edge (which contains the type of both the source and target node),
        - a dictionary mapping each (node or edge) type to another dictionary listing which column is extracted to which property.

    Note that, when using the `configure` mapping,
    types are created by default in the `ontoweaver.types` module,
    so that you may access the list of all declared types by using:
        - `ontoweaver.types.all.nodes()`,
        - `ontoweaver.types.all.node_fields()`,
        - `ontoweaver.types.all.edges()`,
        - `ontoweaver.types.all.edge_fields()`.
    """

    def __init__(self,
                 df: pd.DataFrame,
                 subject_transformer: base.Transformer,
                 transformers: Iterable[base.Transformer],
                 metadata: Optional[dict] = None,
                 validator: Optional[validate.InputValidator] = None,
                 type_affix: Optional[TypeAffixes] = TypeAffixes.suffix,
                 type_affix_sep: Optional[str] = ":",
                 parallel_mapping: int = 0,
                 raise_errors = True,
                 ):
        """
        Instantiate the adapter.

        Args:
            df (pd.DataFrame): The table containing the input data.
            subject_transformer (base.Transformer): The transformer that maps the subject node.
            transformers (Iterable[base.Transformer]): List of transformer instances that map the data frame to nodes and edges.
            metadata (Optional[dict]): Metadata to be added to all the nodes and edges.
            type_affix (Optional[TypeAffixes]): Where to add a type annotation to the labels (either TypeAffixes.prefix, TypeAffixes.suffix or TypeAffixes.none).
            type_affix_sep (Optional[str]): String used to separate a label from the type annotation (WARNING: double-check that your BioCypher config does not use the same character as a separator).
            parallel_mapping (int): Number of workers to use in parallel mapping. Defaults to 0 for sequential processing.
            raise_errors (bool): if True, will raise an exception when an error is encountered, else, will log the error and try to proceed.
        """
        super().__init__(raise_errors)

        logger.info("DataFrame info:")
        # logger.info(df.info())
        logger.debug("Columns:")
        for c in df.columns:
            logger.debug(f"\t`{c}`")
        pd.set_option('display.max_rows', 30)
        pd.set_option('display.max_columns', 30)
        pd.set_option('display.width', 150)
        logger.info("\n" + str(df))
        self.df = df

        self.validator = validator

        if not type_affix in TypeAffixes:
            self.error(f"`type_affix`={type_affix} is not one of the allowed values ({[t for t in TypeAffixes]})", exception = exceptions.ConfigError)
        else:
            self.type_affix = type_affix

        self.type_affix_sep = type_affix_sep

        self.subject_transformer = subject_transformer
        self.transformers = transformers
        self.property_transformers = [] # populated at parsing in self.properties.
        self.metadata = metadata
        # logger.debug(self.target_element_properties)
        self.parallel_mapping = parallel_mapping

    # FIXME not used, maybe will come in handy?
    def source_type(self, row):
        """
        Accessor to the row type actually used by `run`.

        You may overload this function if you want
        to make the row type dependent on some column value.

        By default, just return the default row type defined in the constructor,
        without taking the row values into account.

        Args:
            row: The current row of the DataFrame.

        Returns:
            The row type.
        """
        return self.row_type


    def make_id(self, entry_type, entry_name):
        """
        LabelMaker a unique id for the given cell consisting of the entry name and type,
        taking into account affix and separator configuration.

        Args:
            entry_type: The type of the entry.
            entry_name: The name of the entry.

        Returns:
            str: The created ID.

        Raises:
            ValueError: If the ID creation fails.
        """
        assert(type(entry_type) == str)
        if type(entry_name) != str:
            logger.warning(f"Identifier '{entry_name}' (of type '{entry_type}') is not a string, I had to convert it explicitely, check that the related transformer yields a string.")
            entry_name = str(entry_name)

        if self.type_affix == TypeAffixes.prefix:
            id = f'{entry_type}{self.type_affix_sep}{entry_name}'
        elif self.type_affix == TypeAffixes.suffix:
            id = f'{entry_name}{self.type_affix_sep}{entry_type}'
        elif self.type_affix == TypeAffixes.none:
            id = f'{entry_name}'

        if id:
            logger.debug(f"\t\tFormatted ID `{id}` for cell value `{entry_name}` of type: `{entry_type}`")
            return id
        else:
            self.error(f"Failed to format ID for cell value: `{entry_name}` of type: `{entry_type}`", exception = exceptions.DeclarationError)


    def valid(self, val):
        """
        Checks if cell value is valid - not a `nan`.

        Args:
            val: The value to check.

        Returns:
            bool: True if the value is valid, False otherwise.
        """
        if pd.api.types.is_numeric_dtype(type(val)):
            if (math.isnan(val) or val == float("nan")):
                return False
        elif str(val) == "nan":  # Conversion from Pandas' `object` needs to be explicit.
            return False
        return True


    def properties(self, property_dict, row, i, edge_t, node_t, node = False):

        """
        Extract properties of each property category for the given node type.
        If no properties are found, return an empty dictionary.

        Args:
            property_dict: Dictionary of property mappings.
            row: The current row of the DataFrame.
            i: The index of the current row.
            edge_t: The type of the edge of the current transformer.
            node_t: The type of the node of the current transformer.
            node: True if the object created is a node, False otherwise.

        Returns:
            dict: Extracted properties.
        """
        properties = {}

        for prop_transformer, property_name in property_dict.items():
            for property, none_node, none_edge in prop_transformer(row, i):
                if property:
                    properties[property_name] = str(property).replace("'", "`")
                    logger.info(f"                 {prop_transformer} to property `{property_name}` with value `{properties[property_name]}`.")
                else:
                    self.error(f"Failed to extract valid property with {prop_transformer.__repr__()} for {i}th row.", indent=2, exception = exceptions.TransformerDataError)
                    continue

        # If the metadata dictionary is not empty, add the metadata to the property dictionary.
        if self.metadata:
            if node:
                elem = node_t
            else:
                elem = edge_t
            if elem.__name__ in self.metadata:
                for key, value in self.metadata[elem.__name__].items():
                    properties[key] = value

        return properties


    def make_node(self, node_t, id, properties):
        """
        LabelMaker nodes of a certain type.

        Args:
            node_t: The type of the node.
            id: The ID of the node.
            properties: The properties of the node.

        Returns:
            The created node.
        """
        return node_t(id=id, properties=properties)


    def make_edge(self, edge_t, id_target, id_source, properties):
        """
        LabelMaker edges of a certain type.

        Args:
            edge_t: The type of the edge.
            id_target: The ID of the target node.
            id_source: The ID of the source node.
            properties: The properties of the edge.

        Returns:
            The created edge.
        """
        return edge_t(id_source=id_source, id_target=id_target, properties=properties)

    def run(self):
        """Iterate through dataframe in parallel and map cell values according to YAML file, using a list of transformers."""

        # Thread-safe containers with their respective locks
        self._nodes = []
        self._edges = []
        self._errors = []
        self._nodes_lock = threading.Lock()
        self._edges_lock = threading.Lock()
        self._errors_lock = threading.Lock()
        self._row_lock = threading.Lock()
        self._transformations_lock = threading.Lock()
        self._local_nb_nodes_lock = threading.Lock()

        # Function to process a single row and collect operations
        def process_row(row_data):
            i, row = row_data
            local_nodes = []
            local_edges = []
            local_errors = []
            local_rows = 0
            local_transformations = 0
            local_nb_nodes = 0

            logger.debug(f"Process row {i}...")
            local_rows += 1
            # There can be only one subject, so transformers yielding multiple IDs cannot be used.
            logger.debug("\tLabelMaker subject node:")
            subject_generator_list = list(self.subject_transformer(row, i))
            if (len(subject_generator_list) > 1):
                local_errors.append(self.error(f"You cannot use a transformer yielding multiple IDs as a subject. "
                                               f"Subject Transformer `{self.subject_transformer}` produced multiple IDs: "
                                               f"{subject_generator_list}", indent=2, exception = exceptions.TransformerInterfaceError))

            source_id, subject_edge, subject_node = subject_generator_list[0]

            if self.subject_transformer.final_type:
                # If a final_type attribute is present in the transformer, use it as the source node type, instead
                # of the default type.
                subject_node = self.subject_transformer.final_type

            if source_id:
                source_node_id = self.make_id(subject_node.__name__, source_id)

                if source_node_id:
                    logger.debug(f"\t\tDeclared subject ID: {source_node_id}")
                    local_nodes.append(self.make_node(node_t=subject_node, id=source_node_id,
                                                      # FIXME: Should we use the meta-way of accessing node properties as well?
                                                      # FIXME: This would require a refactoring of the transformer interfaces and tabular.run.
                                                      properties=self.properties(self.subject_transformer.properties_of,
                                                                                 row, i, subject_edge, subject_node,
                                                                                 node=True)))
                else:
                    local_errors.append(self.error(f"Failed to declare subject ID for row #{i}: `{row}`.",
                                                   indent=2, exception = exceptions.DeclarationError))

                # Loop over list of transformer instances and label_maker corresponding nodes and edges.
                # FIXME the transformer variable here shadows the transformer module.
                for j,transformer in enumerate(self.transformers):
                    local_transformations += 1
                    logger.debug(f"\tCalling transformer: {transformer}...")
                    for target_id, target_edge, target_node in transformer(row, i):
                        local_nb_nodes += 1
                        if target_id and target_edge and target_node:

                            if transformer.final_type:
                                # If a final_type attribute is present in the transformer, use it as the target node type, instead
                                # of the default type.
                                target_node = transformer.final_type

                            target_node_id = self.make_id(target_node.__name__, target_id)
                            logger.debug(f"\t\tMake node {target_node_id}")
                            local_nodes.append(self.make_node(node_t=target_node, id=target_node_id,
                                                              # FIXME: Should we use the meta-way of accessing node properties as well?
                                                              # FIXME: This would require a refactoring of the transformer interfaces and tabular.run.
                                                              properties=self.properties(transformer.properties_of, row,
                                                                                         i, target_edge, target_node, node=True)))

                            # If a `from_subject` attribute is present in the transformer, loop over the transformer
                            # list to find the transformer instance mapping to the correct type, and then label_maker new
                            # subject id.

                            # FIXME add hook functions to be overloaded.

                            # FIXME: Make from_subject reference a list of subjects instead of using the add_edge function.

                            if hasattr(transformer, "from_subject"):

                                found_valid_subject = False

                                for t in self.transformers:
                                    if transformer.from_subject == t.target_type:
                                        found_valid_subject = True
                                        for s_id, s_edge, s_node in t(row, i):
                                            if s_id and s_edge and s_node:
                                                if t.final_type:
                                                    s_node = t.final_type
                                                subject_id = s_id
                                                subject_node_id = self.make_id(t.target_type, subject_id)
                                                logger.debug(
                                                    f"\t\tMake edge from {subject_node_id} toward {target_node_id}")
                                                local_edges.append(
                                                    self.make_edge(edge_t=target_edge, id_source=subject_node_id,
                                                                   id_target=target_node_id,
                                                                   properties=self.properties(target_edge.fields(),
                                                                                              row, i, s_edge, s_node)))

                                            else:
                                                local_errors.append(self.error(
                                                    f"No valid identifiers from {t} for {i}th row, when trying to change default subject type",
                                                    f"by {transformer} with `from_subject` attribute.",
                                                    indent=7, section="transformers", index=j,
                                                    exception=exceptions.TransformerDataError))
                                                continue

                                    else:
                                        # The transformer instance type does not match the type in the `from_subject` attribute.
                                        continue

                                if not found_valid_subject:
                                    local_errors.append(self.error(f"\t\t\tInvalid subject declared from {transformer}."
                                                                   f" The subject you declared in the `from_subject` directive: `"
                                                                   f"{transformer.from_subject}` must not be the same as the default subject type.",
                                                                   exception=exceptions.ConfigError))


                            else: # no attribute `from_subject` in `transformer`
                                logger.debug(f"\t\tMake edge {target_edge.__name__} from {source_node_id} toward {target_node_id}")
                                local_edges.append(self.make_edge(edge_t=target_edge, id_target=target_node_id,
                                                                  id_source=source_node_id,
                                                                  properties=self.properties(target_edge.fields(),
                                                                                             row, i, target_edge, target_node)))
                        else:
                            local_errors.append(self.error(f"No valid target node identifier from {transformer}"
                                                           f" for {i}th row.", indent=2, section="transformers",
                                                           index=j, exception = exceptions.TransformerDataError))
                            continue
            else:
                local_errors.append(self.error(f"No valid source node identifier from {self.subject_transformer} for {i}th row."
                                               f" This row will be skipped.", indent=2, exception = exceptions.TransformerDataError))
                return local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes

            return local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes
        # End of process_row local function

        nb_rows = 0
        nb_transformations = 0
        nb_nodes = 0

        if self.parallel_mapping > 0:
            logger.info(f"Processing dataframe in parallel. Number of workers set to: {self.parallel_mapping} ...")
            # Process the dataset in parallel using ThreadPoolExecutor
            with ThreadPoolExecutor() as executor:
                # Map the process_row function across the dataframe
                results = list(executor.map(process_row, self.df.iterrows()))

            # Append the results in a thread-safe manner after all rows have been processed
            for local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes in results:
                with self._nodes_lock:
                    self.nodes_append(local_nodes)
                with self._edges_lock:
                    self.edges_append(local_edges)
                with self._errors_lock:
                    self.errors += local_errors
                with self._row_lock:
                    nb_rows += local_rows
                with self._transformations_lock:
                    nb_transformations += local_transformations
                with self._local_nb_nodes_lock:
                    nb_nodes += local_nb_nodes

        elif self.parallel_mapping == 0:
            logger.info(f"Processing dataframe sequentially...")
            for i, row in self.df.iterrows():
                local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes = process_row((i, row))
                self.nodes_append(local_nodes)
                self.edges_append(local_edges)
                self.errors += local_errors
                nb_rows += local_rows
                nb_transformations += local_transformations
                nb_nodes += local_nb_nodes

        else:
            self.error(f"Invalid value for `parallel_mapping` ({self.parallel_mapping})."
                       f" Pass 0 for sequential processing, or the number of workers for parallel processing.", exception = exceptions.ConfigError)

        # Final logger
        error_count = {}
        for transformer in chain([self.subject_transformer], self.transformers, self.property_transformers):
            if transformer.output_validator:
                for msg,err in transformer.output_validator.messages.items():
                    desc = f"in {err['section']} with {transformer}: {msg}"
                    # FIXME do we want the number of validation errors or the number of bad cell values?
                    # error_count[desc] = error_count.get(desc, 0) + err['count']
                    error_count[desc] = error_count.get(desc, err['count'])
        for desc,count in error_count.items():
            logger.error(f"Recorded {count} times a validation error {desc}")

        if self.errors:
            logger.error(
                f"Recorded {len(self.errors)} errors while processing {nb_transformations} transformations with {1+len(self.transformers)} node transformers, producing {nb_nodes} nodes for {nb_rows} rows.")
            # logger.debug("\n".join(self.errors))
        else:
            logger.info(
                f"Performed {nb_transformations} transformations with {1+len(self.transformers)} node transformers, producing {nb_nodes} nodes for {nb_rows} rows.")


def extract_table(df: pd.DataFrame, config: dict, parallel_mapping = 0, module = types, affix = "suffix", separator = ":", validate_output = False, raise_errors = True):
    """
    Proxy function for extracting from a table all nodes, edges and properties
    that are defined in a PandasAdapter configuration.

    Args:
        df (pd.DataFrame): The DataFrame containing the input data.
        config (dict): The configuration dictionary.
        parallel_mapping (int): Number of workers to use in parallel mapping. Defaults to 0 for sequential processing.
        module: The module in which to insert the types declared by the configuration.
        affix (str): The type affix to use (default is "suffix").
        separator (str): The separator to use between labels and type annotations (default is ":").
        validate_output: Whether to validate the output of the transformers. Defaults to False.
        raise_errors: Whether to raise errors encountered during the mapping, and stop the mapping process. Defaults to True.

    Returns:
        PandasAdapter: The configured adapter.
    """
    parser = YamlParser(config, module, validate_output = validate_output, raise_errors = raise_errors)
    mapping = parser()

    adapter = PandasAdapter(
        df,
        *mapping,
        type_affix=affix,
        type_affix_sep=separator,
        parallel_mapping=parallel_mapping,
        raise_errors = raise_errors,
    )

    adapter.run()

    return adapter



class Declare(errormanager.ErrorManager):
    """
    Declarations of functions used to declare and instantiate object classes used by the Adapter for the mapping
    of the data frame.

    Args:
        module: The module in which to insert the types declared by the configuration.
    """

    def __init__(self,
                 module=types,
                 raise_errors = True,
                 ):
        super().__init__(raise_errors)
        self.module = module

    def make_node_class(self, name, properties={}, base=base.Node):
        """
        LabelMaker a node class with the given name and properties.

        Args:
            name: The name of the node class.
            properties (dict): The properties of the node class.
            base: The base class for the node class.

        Returns:
            The created node class.
        """
        # If type already exists, return it.
        if hasattr(self.module, name):
            cls = getattr(self.module, name)
            logger.debug(
                f"\t\tNode class `{name}` (prop: `{cls.fields()}`) already exists, I will not create another one.")
            for p in properties.values():
                if p not in cls.fields():
                    logger.warning(f"\t\t\tProperty `{p}` not found in declared fields for node class `{cls.__name__}`.")
            return cls

        def fields():
            return list(properties.values())

        attrs = {
            "__module__": self.module.__name__,
            "fields": staticmethod(fields),
        }
        t = pytypes.new_class(name, (base,), {}, lambda ns: ns.update(attrs))
        logger.debug(f"\t\tDeclare Node class `{t.__name__}` (prop: `{properties}`).")
        setattr(self.module, t.__name__, t)
        return t

    def make_edge_class(self, name, source_t, target_t, properties={}, base=base.Edge):
        """
        LabelMaker an edge class with the given name, source type, target type, and properties.

        Args:
            name: The name of the edge class.
            source_t: The source type of the edge.
            target_t: The target type of the edge.
            properties (dict): The properties of the edge class.
            base: The base class for the edge class.

        Returns:
            The created edge class.
        """
        # If type already exists, check if the fields are the same.
        if hasattr(self.module, name):
            cls = getattr(self.module, name)
            cls_fields = cls.fields()

            # Compare the properties with the existing class fields
            if properties == cls_fields:
                logger.info(
                    f"\t\tEdge class `{name}` (prop: `{cls_fields}`) already exists with the same properties, I will not create another one.")
                return cls

            logger.warning(f"\t\tEdge class `{name}` already exists, but properties do not match.")
            # If properties do not match, we proceed to create a new class with the new properties.

        def fields():
            return properties

        def st():
            return source_t

        def tt():
            return [target_t]

        attrs = {
            "__module__": self.module.__name__,
            "fields": staticmethod(fields),
            "source_type": staticmethod(st),
            "target_type": staticmethod(tt),
        }
        t = pytypes.new_class(name, (base,), {}, lambda ns: ns.update(attrs))
        logger.debug(f"\t\tDeclare Edge class `{t.__name__}` (prop: `{properties}`).")
        setattr(self.module, t.__name__, t)
        return t

    def make_transformer_class(self, transformer_type, multi_type_dictionary = None, branching_properties = None,
                               properties=None, columns=None, output_validator=None, label_maker = None, raise_errors = True, **kwargs):
        """
        LabelMaker a transformer class with the given parameters.

        Args:
            multi_type_dictionary: Dictionary of regex rules and corresponding types in case of cell value match.
            transformer_type: The class of the transformer.
            properties: The properties of the transformer.
            columns: The columns to be processed by the transformer.
            output_validator: validate.OutputValidator instance for transformer output validation.
            raise_errors: Whether to raise errors encountered during the mapping, and stop the mapping process. Defaults to True.
            **kwargs: Additional keyword arguments.

        Returns:
            The created transformer class.

        Raises:
            TypeError: If the transformer type is not an existing transformer.
        """
        if hasattr(transformer, transformer_type):
            parent_t = getattr(transformer, transformer_type)
            kwargs.setdefault("subclass", parent_t)
            if not issubclass(parent_t, base.Transformer):
                self.error(f"Object `{transformer_type}` is not an existing transformer.", exception = exceptions.DeclarationError)
                logger.debug(f"\t\tDeclare Transformer class '{transformer_type}' for node type '{nt}'")
            return parent_t(properties_of=properties,
                            columns=columns,
                            output_validator=output_validator,
                            multi_type_dict = multi_type_dictionary,
                            branching_properties = branching_properties,
                            label_maker = label_maker,
                            raise_errors = raise_errors,
                            **kwargs)
        else:
            # logger.debug(dir(generators))
            self.error(f"Cannot find a transformer class with name `{transformer_type}`.", exception = exceptions.DeclarationError)

class YamlParser(Declare):
    """
    Parse a table extraction configuration and return the three objects needed to configure an Adapter.

    The config is a dictionary containing only strings, as converted from the following YAML description:

    .. code-block:: yaml

            row:
               map:
                  columns:
                    - <MY_COLUMN_NAME>
                  to_subject: <MY_SUBJECT_TYPE>
            transformers:
                - map:
                    columns:
                        - <MY_COLUMN_NAME>
                    to_object: <MY_OBJECT_TYPE>
                    via_relation: <MY_RELATION_TYPE>
                - map:
                    columns:
                        - <MY_OTHER_COLUMN>
                    to_property:
                        - <MY_PROPERTY>
                    for_objects:
                        - <MY_OBJECT_TYPE>

    This maps the table row to a MY_SUBJECT_TYPE node type, adding an edge of type MY_RELATION_TYPE,
    between the MY_SUBJECT_TYPE node and another MY_OBJECT_TYPE node. The data in MY_OTHER_COLUMN is mapped
    to the MY_PROPERTY property of the MY_OBJECT_TYPE node. Note that `to_properties` may effectively map to
    an edge type or several types.

    In order to allow the user to write mappings configurations using their preferred vocabulary, the following
    keywords are interchangeable:
        - subject = row = entry = line,
        - columns = fields,
        - to_target = to_object = to_node
        - via_edge = via_relation = via_predicate.

    :param dict config: A configuration dictionary.
    :param module: The module in which to insert the types declared by the configuration.
    :return tuple: subject_transformer, transformers, metadata as needed by the Adapter.
    """

    def __init__(self, config: dict, module=types, validate_output = False, raise_errors = True):
        """
        Initialize the YamlParser.

        Args:
            config (dict): The configuration dictionary.
            validate_output (bool): Whether to validate the output of the transformers. Defaults to False.
            module: The module in which to insert the types declared by the configuration.
        """
        super().__init__(module, raise_errors = raise_errors)
        self.config = config
        self.validate_output = validate_output


        logger.debug(f"Classes will be created in module '{self.module}'")

    def get_not(self, keys, pconfig=None):
        """
        Get the first dictionary (key, item) not matching any of the passed keys.

        Args:
            keys: The keys to exclude.
            pconfig: The configuration dictionary to search in (default is self.config).

        Returns:
            dict: The first dictionary not matching any of the passed keys.
        """
        res = {}
        if not pconfig:
            pconfig = self.config
        for k in pconfig:
            if k not in keys:
                res[k] = pconfig[k]
        return res

    def get(self, keys, pconfig=None):
        """
        Get a dictionary item matching any of the passed keys.

        Args:
            keys: The keys to search for.
            pconfig: The configuration dictionary to search in (default is self.config).

        Returns:
            The first item matching any of the passed keys, or None if no match is found.
        """
        if not pconfig:
            pconfig = self.config
        for k in keys:
            if k in pconfig:
                return pconfig[k]
        return None

    def _extract_metadata(self, k_metadata_column, metadata_list, metadata, types, columns):
        """
        Extract metadata and update the metadata dictionary.

        Args:
            k_metadata_column (list): List of keys to be used for adding source column names.
            metadata_list (list): List of metadata items to be added.
            metadata (dict): The metadata dictionary to be updated.
            types (str): The type of the node or edge.
            columns (list): List of columns to be added to the metadata.

        Returns:
            dict: The updated metadata dictionary.
        """
        # TODO: Redundant code with the _extract_metadata function.
        if metadata_list and types:
            if type(types) != set:
                t = types
                metadata.setdefault(t, {})
                for item in metadata_list:
                    metadata[t].update(item)
                for key in k_metadata_column:
                    if key in metadata[t]:
                        # Use the value of k_metadata_column as the key.
                        key_name = metadata[t][key]
                        # Remove the k_metadata_column key from the metadata dictionary.
                        if key_name in metadata[t]:
                            msg = f"The key you used for adding source column names: `{key_name}` to node: `{t}` already exists in the metadata dictionary."
                            # FIXME is it an error or a warning?
                            # self.error(msg)
                            logger.warning(msg)
                        del metadata[t][key]
                        if columns:
                            # TODO make the separator a parameter.
                            metadata[t][key_name] = ", ".join(columns)
            else:
                for t in types:
                    metadata.setdefault(t, {})
                    for item in metadata_list:
                        metadata[t].update(item)
                    for key in k_metadata_column:
                        if key in metadata[t]:
                            # Use the value of k_metadata_column as the key.
                            key_name = metadata[t][key]
                            # Remove the k_metadata_column key from the metadata dictionary.
                            if key_name in metadata[t]:
                                msg = f"The key you used for adding source column names: `{key_name}` to node: `{t}` already exists in the metadata dictionary."
                                # FIXME is it an error or a warning?
                                # self.error(msg)
                                logger.warning(msg)
                            del metadata[t][key]
                            if columns:
                                # TODO make the separator a parameter.
                                metadata[t][key_name] = ", ".join(columns)

            return metadata
        else:
            return None


    def _get_input_validation_rules(self,):
        """
        Extract input data validation schema from yaml file and instantiate a Pandera DataFrameSchema object and validator.
        """
        k_validate = ["validate"]
        validation_rules = self.get(k_validate)
        yaml_validation_rules = yaml.dump(validation_rules, default_flow_style=False)
        validator = None

        try:
            validation_schema = pa.DataFrameSchema.from_yaml(yaml_validation_rules)
            validator = validate.InputValidator(validation_schema, raise_errors=self.raise_errors)
        except Exception as e:
            self.error(f"Failed to parse the input validation schema: {e}", exception=exceptions.ConfigError)

        return validator

    def _make_output_validator(self, output_validation_rules = None):
        """
        LabelMaker a validator for the output of a transformer.

        Args:
            output_validation_rules: The output validation rules for the transformer extracted from yaml file.

        Returns:
            validate.OutputValidator: The created validator.
        """

        if self.validate_output:
            if output_validation_rules:
                output_validator = validate.OutputValidator(raise_errors=self.raise_errors)
                # Adjust the formatting for output validation rules to match the expected format. This is so the
                # user would not have to type `columns` and `cell_value` in the configuration file each time.
                dict_output_validation_rules = {"columns": {"cell_value": output_validation_rules}}
                yaml_output_validation_rules = yaml.dump(dict_output_validation_rules, default_flow_style=False)
                output_validator.update_rules(pa.DataFrameSchema.from_yaml(yaml_output_validation_rules))
            else:
                output_validator = validate.SimpleOutputValidator(raise_errors=self.raise_errors)
        else:
            output_validator = validate.SkipValidator(raise_errors=self.raise_errors)

        return output_validator

    def _extract_final_type_class(self, final_type, possible_types, metadata, metadata_list, columns, properties_of):
        """
        Extract metadata and class for the final type and update the metadata dictionary.
        """

        if final_type:
            final_type_class = self.make_node_class(final_type, properties_of.get(final_type, {}))
            possible_types.add(final_type_class.__name__)
            extracted_s_final_type_metadata = self._extract_metadata(self.k_metadata_column,
                                                                     metadata_list, metadata,
                                                                     final_type,
                                                                     columns)
            if extracted_s_final_type_metadata:
                metadata.update(extracted_s_final_type_metadata)

            return final_type_class

        return None


    def _make_branching_dict(self, subject: bool, match_parser, properties_of, metadata_list, metadata, columns, final_type_class,
                             multi_type_dictionary, possible_node_types, possible_edge_types):
        """
        Helper function to parse the `match` clause of the YAML configuration file for subject and target transformers.
        """

        k_extract  = self.k_subject if subject else self.k_target

        for entry in match_parser:
            for k, v in entry.items():
                if isinstance(v, dict):
                    key = k
                    multi_type_dictionary[key] = {k1: v1 for k1, v1 in v.items()}
                    alt_type = self.get(k_extract, v)
                    alt_type_class = self.make_node_class(alt_type, properties_of.get(alt_type, {}))

                    possible_node_types.add(alt_type)

                    alt_final_type = self.get(self.k_final_type, v)

                    alt_final_type_class = self._extract_final_type_class(alt_final_type, possible_node_types, metadata,
                                                                          metadata_list, columns, properties_of)

                    if not subject:
                        alt_edge = self.get(self.k_edge, v)
                        alt_edge_class = self.make_edge_class(alt_edge, None, alt_type_class,
                                                              properties_of.get(alt_edge, {}))

                        possible_edge_types.add(alt_edge)

                        extracted_alt_edge_metadata = self._extract_metadata(self.k_metadata_column,
                                                                             metadata_list, metadata, alt_edge,
                                                                             None)

                        if extracted_alt_edge_metadata:
                            metadata.update(extracted_alt_edge_metadata)

                    extracted_alt_type_metadata = self._extract_metadata(self.k_metadata_column,
                                                                           metadata_list, metadata, alt_type,
                                                                           columns)
                    if extracted_alt_type_metadata:
                        metadata.update(extracted_alt_type_metadata)

                    if extracted_alt_type_metadata:
                        metadata.update(extracted_alt_type_metadata)

                    multi_type_dictionary[key] = {
                        'to_object': alt_type_class,
                        # Via relation is always None for subject, since there is never an edge declared for the subject type.
                        'via_relation': None if subject is True else alt_edge_class,
                        # We first try and declare the final type passed to the whole class of the
                        # transformer, and if it is not defined, we use the final type of the
                        # alternative node type, under the `match` clause.
                        'final_type': final_type_class if final_type_class else alt_final_type_class
                    }

    def parse_properties(self, properties_of, possible_subject_types, transformers_list):
        """
        Parse the properties of the transformers defined in the YAML mapping, and update the properties_of dictionary.
        """
        logger.debug(f"Parse properties...")
        for n_transformer, transformer_types in enumerate(transformers_list):
            for transformer_type, field_dict in transformer_types.items():
                if not field_dict:
                    self.error(f"There is no field for the {n_transformer}th transformer: '{transformer_type}',"
                               f" did you forget an indentation?", "transformers",
                               n_transformer, exception=exceptions.MissingFieldError)

                if any(field in field_dict for field in self.k_properties):
                    object_types = self.get(self.k_prop_to_object, pconfig=field_dict)
                    property_names = self.get(self.k_properties, pconfig=field_dict)
                    if type(property_names) != list:
                        logger.debug(f"\tDeclared singular property")
                        assert (type(property_names) == str)
                        property_names = [property_names]
                    if not object_types:  # FIXME: Creates errors with branching subject types and subject final type features.
                        logger.info(
                            f"No `for_objects` defined for properties {property_names}, I will attach those properties to the row subject(s) `{possible_subject_types}`")
                        if type(possible_subject_types) == set and len(possible_subject_types) == 1:
                            object_types = list(possible_subject_types)[0]
                        if type(possible_subject_types) == set and len(possible_subject_types) > 1:
                            object_types = list(possible_subject_types)
                    if type(object_types) != list:
                        logger.debug(f"\tDeclared singular for_object: `{object_types}`")
                        assert (type(object_types) == str)
                        object_types = [object_types]

                    column_names = self.get(self.k_columns, pconfig=field_dict)
                    if column_names != None and type(column_names) != list:
                        logger.debug(f"\tDeclared singular column `{column_names}`")
                        assert (type(column_names) == str)
                        column_names = [column_names]
                    gen_data = self.get_not(self.k_target + self.k_edge + self.k_columns, pconfig=field_dict)

                    # Parse the validation rules for the output of the property transformer.
                    p_output_validation_rules = self.get(self.k_validate_output, pconfig=field_dict)
                    p_output_validator = self._make_output_validator(p_output_validation_rules)

                    prop_transformer = self.make_transformer_class(transformer_type, columns=column_names,
                                                                   output_validator=p_output_validator,
                                                                   label_maker=make_labels.SimpleLabelMaker(
                                                                       raise_errors=self.raise_errors), **gen_data)

                    for object_type in object_types:
                        properties_of.setdefault(object_type, {})
                        for property_name in property_names:
                            properties_of[object_type].setdefault(prop_transformer, property_name)
                        logger.debug(f"\t\tDeclared property mapping for `{object_type}`: {properties_of[object_type]}")

        return properties_of

    def parse_subject(self, properties_of, transformers_list, metadata_list, metadata):
        """
        Parse the subject transformer and its properties from the YAML mapping.
        """

        logger.debug(f"Declare subject type...")
        subject_transformer_dict = self.get(self.k_row)
        subject_transformer_class = list(subject_transformer_dict.keys())[0]
        subject_kwargs = self.get_not(self.k_subject_type + self.k_columns, subject_transformer_dict[
            subject_transformer_class])  # FIXME shows redundant information filter out the keys that are not needed.
        subject_columns = self.get(self.k_columns, subject_transformer_dict[subject_transformer_class])
        subject_final_type = self.get(self.k_final_type, subject_transformer_dict[subject_transformer_class])
        if subject_columns != None and type(subject_columns) != list:
            logger.debug(f"\tDeclared singular subject’s column `{subject_columns}`")
            assert (type(subject_columns) == str)
            subject_columns = [subject_columns]

        # Parse the validation rules for the output of the subject transformer.
        subject_output_validation_rules = self.get(self.k_validate_output,
                                                   subject_transformer_dict[subject_transformer_class])
        subject_output_validator = self._make_output_validator(subject_output_validation_rules)

        subject_multi_type_dict = {}
        # TODO: assert subject_transformer_dict contains only ONE transformer.

        subject_transformer_params = subject_transformer_dict[subject_transformer_class]

        possible_subject_types = set()
        subject_branching = False

        s_final_type_class = self._extract_final_type_class(subject_final_type, possible_subject_types, metadata,
                                                            metadata_list, subject_columns, properties_of)

        if "match" in subject_transformer_params:

            subject_branching = True

            self._make_branching_dict(subject = True, match_parser = subject_transformer_params["match"],
                                      properties_of = properties_of, metadata_list = metadata_list, metadata = metadata,
                                      columns = subject_columns, final_type_class = s_final_type_class,
                                      multi_type_dictionary = subject_multi_type_dict, possible_node_types = possible_subject_types,
                                      possible_edge_types = set())

            source_t = None # Subject_type declared None because the subject transformer is a branching transformer.
            subject_type = None
            logger.debug(f"Parse subject transformer...")

            if "match_type_from_column" in subject_kwargs: #FIXME should be a k_variable just like the others.
                s_label_maker = make_labels.MultiTypeOnColumnLabelMaker(raise_errors=self.raise_errors,
                                                                        match_type_from_column=subject_kwargs['match_type_from_column'])
            else:
                s_label_maker = make_labels.MultiTypeLabelMaker(raise_errors=self.raise_errors)

        # "None" key is used to return any type of string, in case no branching is needed.
        else:
            subject_type = self.get(self.k_subject_type, subject_transformer_dict[subject_transformer_class])
            source_t = self.make_node_class(subject_type, properties_of.get(subject_type, {}))

            subject_multi_type_dict = {'None': {
                'to_object': source_t,
                'via_relation': None,
                'final_type': s_final_type_class
            }}

            possible_subject_types.add(source_t.__name__)

            s_label_maker = make_labels.SimpleLabelMaker(raise_errors=self.raise_errors)

        properties_of = self.parse_properties(properties_of, possible_subject_types, transformers_list)

        subject_transformer = self.make_transformer_class(transformer_type=subject_transformer_class,
                                                          multi_type_dictionary=subject_multi_type_dict,
                                                          branching_properties=properties_of if subject_branching else None,
                                                          properties=properties_of.get(subject_type,
                                                                                       {}) if not subject_branching else None,
                                                          columns=subject_columns,
                                                          output_validator=subject_output_validator,
                                                          label_maker=s_label_maker, raise_errors=self.raise_errors,
                                                          **subject_kwargs)

        logger.debug(f"\tDeclared subject transformer: {subject_transformer}")

        logger.debug(
            f"\tDeclare subject of possible types: '{possible_subject_types}', subject transformer: '{subject_transformer_class}', "
            f"subject kwargs: '{subject_kwargs}', subject columns: '{subject_columns}'")

        extracted_metadata = self._extract_metadata(self.k_metadata_column, metadata_list, metadata, possible_subject_types, subject_columns)
        if extracted_metadata:
            metadata.update(extracted_metadata)

        return possible_subject_types, subject_transformer, source_t, subject_columns


    def parse_targets(self, transformers_list, properties_of, source_t, metadata_list, metadata):
        """
        Parse the target transformers and their properties from the YAML mapping.
        """

        transformers = []
        possible_target_types = set()
        possible_edge_types = set()

        logger.debug(f"Declare types...")
        for n_transformer,transformer_types in enumerate(transformers_list):
            for transformer_type, field_dict in transformer_types.items():
                if not field_dict:
                    continue
                elif any(field in field_dict for field in self.k_properties):
                    if any(field in field_dict for field in self.k_target):
                        prop = self.get(self.k_properties, field_dict)
                        target = self.get(self.k_target, field_dict)
                        self.error(f"ERROR in transformer '{transformer_type}': one cannot "
                                      f"declare a mapping to both properties '{prop}' and object type '{target}'.", "transformers",
                                   n_transformer, exception = exceptions.CardinalityError)
                    continue
                else:
                    if type(field_dict) != dict:
                        self.error(str(field_dict)+" is not a dictionary", exception = exceptions.ParsingDeclarationsError)

                    columns = self.get(self.k_columns, pconfig=field_dict)
                    if type(columns) != list:
                        logger.debug(f"\tDeclared singular column")
                        # The rowIndex transformer is a special case, where the column does not need to be defined in the mapping.
                        # FIXME: In next refactoring do not assert `rowIndex` transformer name, in order to have a generic implementation. (ref: https://github.com/oncodash/ontoweaver/pull/153)
                        if transformer_type != "rowIndex":
                            assert(type(columns) == str)
                            columns = [columns]

                    target = self.get(self.k_target, pconfig=field_dict)
                    if type(target) == list:
                        self.error(f"You cannot declare multiple objects in transformers. For transformer `{transformer_type}`.",
                                   section="transformers", index=n_transformer, indent=1, exception = exceptions.CardinalityError)

                    subject = self.get(self.k_subject, pconfig=field_dict)
                    if type(subject) == list:
                        self.error(f"You cannot declare multiple subjects in transformers. For transformer `{transformer_type}`.",
                                   section="transformers", index=n_transformer, indent=1, exception = exceptions.CardinalityError)

                    edge = self.get(self.k_edge, pconfig=field_dict)
                    if type(edge) == list:
                        self.error(f"You cannot declare multiple relations in transformers. For transformer `{transformer_type}`.",
                                   section="transformers", index=n_transformer, indent=1, exception = exceptions.CardinalityError)

                    gen_data = self.get_not(self.k_target + self.k_edge + self.k_columns + self.k_final_type, pconfig=field_dict)

                    # Extract the final type if defined in the mapping.
                    final_type = self.get(self.k_final_type, pconfig=field_dict)

                    final_type_class = self._extract_final_type_class(final_type, possible_target_types, metadata,
                                                                      metadata_list, columns, properties_of)

                    # Harmonize the use of the `from_subject` and `from_source` synonyms in the configuration, because
                    # from_subject` is used in the transformer class to refer to the source node type.
                    if 'from_source' in gen_data:
                        gen_data['from_subject'] = gen_data['from_source']
                        del gen_data['from_source']

                    multi_type_dictionary = {}

                    if "match" in gen_data:

                        self._make_branching_dict(subject=False, match_parser=gen_data["match"],
                                                  properties_of=properties_of, metadata_list=metadata_list,
                                                  metadata=metadata,
                                                  columns=columns, final_type_class=final_type_class,
                                                  multi_type_dictionary=multi_type_dictionary,
                                                  possible_node_types=possible_target_types,
                                                  possible_edge_types=possible_edge_types)

                    # Parse the validation rules for the output of the transformer. Each transformer gets its own
                    # instance of the OutputValidator with (at least) the default output validation rules.
                    output_validation_rules = self.get(self.k_validate_output, pconfig=field_dict)
                    output_validator = self._make_output_validator(output_validation_rules)

                    if target and edge:
                        logger.debug(f"\tDeclare node .target for `{target}`...")
                        target_t = self.make_node_class(target, properties_of.get(target, {}))
                        possible_target_types.add(target)
                        logger.debug(f"\t\tDeclared target for `{target}`: {target_t.__name__}")
                        if subject:
                            logger.debug(f"\tDeclare subject for `{subject}`...")
                            subject_t = self.make_node_class(subject, properties_of.get(subject, {}))
                            possible_target_types.add(subject_t.__name__)
                            edge_t = self.make_edge_class(edge, subject_t, target_t, properties_of.get(edge, {}))
                            possible_edge_types.add(edge_t.__name__)
                        else:
                            logger.debug(f"\tDeclare edge for `{edge}`...")
                            edge_t = self.make_edge_class(edge, source_t, target_t, properties_of.get(edge, {}))
                            possible_edge_types.add(edge_t.__name__)

                        # "None" key is used to return any type of string, in case no branching is needed.
                        multi_type_dictionary['None'] = {
                            'to_object': target_t,
                            'via_relation': edge_t,
                            'final_type': final_type_class
                        }

                        # Parse the validation rules for the output of the transformer. Each transformer gets its own
                        # instance of the OutputValidator with (at least) the default output validation rules.
                        output_validation_rules = self.get(self.k_validate_output, pconfig=field_dict)
                        output_validator = self._make_output_validator(output_validation_rules)

                        logger.debug(f"\tDeclare transformer `{transformer_type}`...")
                        target_transformer = self.make_transformer_class(transformer_type=transformer_type,
                                                                         multi_type_dictionary=multi_type_dictionary,
                                                                         properties=properties_of.get(target, {}),
                                                                         columns=columns,
                                                                         output_validator=output_validator,
                                                                         label_maker=make_labels.SimpleLabelMaker(
                                                                             raise_errors=self.raise_errors),
                                                                         raise_errors=self.raise_errors, **gen_data)
                        transformers.append(target_transformer)
                        logger.debug(f"\t\tDeclared mapping `{columns}` => `{edge_t.__name__}`")
                    elif (target and not edge) or (edge and not target):
                        self.error(f"Cannot declare the mapping  `{columns}` => `{edge}` (target: `{target}`), "
                                   f"missing either an object or a relation.", "transformers", n_transformer,
                                   indent=2, exception = exceptions.MissingDataError)


                    elif multi_type_dictionary and "match_type_from_column" in gen_data and not target and not edge:
                        target_transformer = self.make_transformer_class(transformer_type=transformer_type,
                                                                         multi_type_dictionary=multi_type_dictionary,
                                                                         branching_properties=properties_of,
                                                                         columns=columns,
                                                                         output_validator=output_validator,
                                                                         label_maker=make_labels.MultiTypeOnColumnLabelMaker(
                                                                             raise_errors=self.raise_errors, match_type_from_column=gen_data["match_type_from_column"]),
                                                                         raise_errors=self.raise_errors, **gen_data)

                        transformers.append(target_transformer)

                    elif multi_type_dictionary and not target and not edge and "match_type_from_column" not in gen_data:
                        target_transformer = self.make_transformer_class(transformer_type=transformer_type,
                                                                         multi_type_dictionary=multi_type_dictionary,
                                                                         branching_properties=properties_of,
                                                                         columns=columns,
                                                                         output_validator=output_validator,
                                                                         label_maker=make_labels.MultiTypeLabelMaker(
                                                                             raise_errors=self.raise_errors),
                                                                         raise_errors=self.raise_errors, **gen_data)

                        transformers.append(target_transformer)

                    # Declare the metadata for the target and edge types.

                    extracted_metadata = self._extract_metadata(self.k_metadata_column, metadata_list, metadata, target, columns)
                    if extracted_metadata:
                        metadata.update(extracted_metadata)

                    if edge:
                        extracted_metadata = self._extract_metadata(self.k_metadata_column, metadata_list, metadata, edge, None)
                        if extracted_metadata:
                            metadata.update(extracted_metadata)

        return transformers, possible_target_types, possible_edge_types


    def __call__(self):
        """
        Parse the configuration and return the subject transformer and transformers.

        Returns:
            tuple: The subject transformer and a list of transformers.
        """
        logger.debug(f"Parse mapping:")

        properties_of = {}
        metadata = {}

        # Various keys are allowed in the config to allow the user to use their favorite ontology vocabulary.
        self.k_row = ["row", "entry", "line", "subject", "source"]
        self.k_subject_type = ["to_subject", "to_object', 'to_node", "to_label", "to_type", "id_from_column"]
        self.k_columns = ["columns", "fields", "column", "match_column", "id_from_column"]
        self.k_target = ["to_target", "to_object", "to_node", "to_label", "to_type"]
        self.k_subject = ["from_subject", "from_source", "to_subject", "to_source", "to_node", "to_label", "to_type"]
        self.k_edge = ["via_edge", "via_relation", "via_predicate"]
        self.k_properties = ["to_properties", "to_property"]
        self.k_prop_to_object = ["for_objects", "for_object"]
        self.k_transformer = ["transformers"]
        self.k_metadata = ["metadata"]
        self.k_metadata_column = ["add_source_column_names_as"]
        self.k_validate_output = ["validate_output"]
        self.k_final_type = ["final_type", "final_object", "final_node", "final_subject", "final_label", "final_target"]

        #TODO create make node class for nested final_type instantiation

        # Extract transformer list and metadata list from the config.
        transformers_list = self.get(self.k_transformer)
        metadata_list = self.get(self.k_metadata)

        # Parse subject type, metadata for subject, and properties for both subject and target types (parse_subject calls parse_properties).
        possible_subject_types, subject_transformer, source_t, subject_columns = self.parse_subject(properties_of, transformers_list, metadata_list, metadata)

        # Parse the target types and target metadata.
        transformers, possible_target_types, possible_edge_types = self.parse_targets(transformers_list, properties_of, source_t, metadata_list, metadata)

        validator = self._get_input_validation_rules()

        if source_t:
            logger.debug(f"source class: {source_t}")
        elif possible_edge_types:
            logger.debug(f"possible_source_types: {possible_subject_types}")
        logger.debug(f"properties_of: {properties_of}")
        logger.debug(f"transformers: {transformers}")
        logger.debug(f"metadata: {metadata}")
        return subject_transformer, transformers, metadata, validator




import re
import sys
import logging
import pandas as pd

from . import base
from. import exceptions
from . import validate
from . import make_value

logger = logging.getLogger("ontoweaver")


def register(transformer_class):
    """Adds the given transformer class to those available to OntoWeaver.

    The given class should inherit from ontoweaver.base.Transformer

    Example::

        import ontoweaver

        class user_transformer(ontoweaver.base.Transformer):
            def __init__(self, target, target_element_properties, edge=None, columns=None, **kwargs):
                super().__init__(target, target_element_properties, edge, columns, **kwargs)

            def __call__(self, row, i):
                for key in self.columns:
                    yield str(row[key])

        ontoweaver.transformer.register( user_transformer )

        # The mapping can now use "user_transformer" in the transformers list.

    Args:
        transformer_class: The class to add to the ontoweaver.transformer module.
    """

    if not issubclass(transformer_class, base.Transformer):
        logging.error(f"{transformer_class.__name__} should inherit from ontoweaver.base.Transformer.", section="transformer.register", exception = exceptions.InterfaceInheritanceError)
    current = sys.modules[__name__]
    setattr(current, transformer_class.__name__, transformer_class)


# NOTE: transformers pass all kwargs to superclass to allow it to show
#       the (additional) user-defined arguments when calling __repr__.


class split(base.Transformer):
    """Transformer subclass used to split cell values at defined separator and label_maker nodes with
    their respective values as id."""

    class ValueMaker(make_value.ValueMaker):

        def __init__(self, raise_errors: bool = True, separator: str = None):
            self.separator = separator
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):

            for key in columns:
                items = str(row[key]).split(self.separator)
                for item in items:
                    yield item

    def __init__(self, properties_of, label_maker = None, branching_properties = None, columns=None, output_validator: validate.OutputValidator = None, raise_errors = True, separator = None, **kwargs):
        """
        Initialize the split transformer.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            sep: Character(s) to use for splitting.
            output_validator: the OutputValidator object used for validating transformer output.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
            separator: The character(s) to use for splitting the cell values. Defaults to ",".
        """

        self.separator = separator

        self.value_maker = self.ValueMaker(raise_errors=raise_errors, separator=self.separator)

        super().__init__(properties_of, self.value_maker, label_maker, branching_properties, columns, output_validator,
                         raise_errors=raise_errors, **kwargs)


    def __call__(self, row, i):
        """
        Process a row and yield split items as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: Each split item from the cell value.
        """
        for value in self.value_maker(self.columns, row, i):
            yield self.create(value, row)


class cat(base.Transformer):
    """Transformer subclass used to concatenate cell values of defined columns and label_maker nodes with
    their respective values as id."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            formatted_items = ""

            for key in columns:
                formatted_items += str(row[key])
                yield formatted_items

    def __init__(self, properties_of, label_maker = None, branching_properties = None, columns=None, output_validator: validate.OutputValidator = None, multi_type_dict = None, raise_errors = True, **kwargs):
        """
        Initialize the cat transformer.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            output_validator: the OutputValidator object used for validating transformer output.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
        """

        self.value_maker = self.ValueMaker(raise_errors=raise_errors)

        super().__init__(properties_of, self.value_maker, label_maker, branching_properties, columns, output_validator,
                         multi_type_dict, raise_errors=raise_errors, **kwargs)

    def __call__(self, row, i):
        """
        Process a row and yield concatenated items as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The concatenated string from the cell values.
        """
        if not self.columns:
            self.error(f"No column declared for the {type(self).__name__} transformer, did you forgot to add a `columns` keyword?", section="cat.call", exception = exceptions.TransformerInputError)

        for value in self.value_maker(self.columns, row, i):
            yield self.create(value, row)




class cat_format(base.Transformer):
    """Transformer subclass used to concatenate cell values of defined columns and label_maker nodes with
    their respective values as id."""

    class ValueMaker(make_value.ValueMaker):

        def __init__(self, raise_errors: bool = True, format_string: str = None):
            self.format_string = format_string
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):

            formatted_string = self.format_string.format_map(row)
            yield formatted_string

    def __init__(self, properties_of, label_maker = None, branching_properties = None, columns=None, output_validator: validate.OutputValidator = None, multi_type_dict = None, raise_errors = True, format_string = None,  **kwargs):
        """
        Initialize the cat_format transformer.

        Args:.
            target_element_properties: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            format_string: A format string containing the column names to assemble.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """

        if not format_string:  # Neither empty string nor None.
            self.error(f"The `format_string` parameter of the `{self.__name__}` transformer cannot be an empty string.")
        self.format_string = format_string
        self.value_maker = self.ValueMaker(raise_errors=raise_errors, format_string=self.format_string)

        super().__init__(properties_of, self.value_maker, label_maker, branching_properties, columns, output_validator, multi_type_dict,
                         raise_errors=raise_errors, **kwargs)

    def __call__(self, row, i):
        """
        Process a row and yield a formatted string as node ID.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The formatted string from the cell values.

        Raises:
            Exception: If the format string is not defined or if invalid content is encountered.
        """

        for value in self.value_maker(self.columns, row, i):
            yield self.create(value, row)

class rowIndex(base.Transformer):
    """Transformer subclass used for the simple mapping of nodes with row index values as id."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            yield i

    def __init__(self, properties_of, label_maker = None, branching_properties = None, columns=None, output_validator: validate.OutputValidator = None, multi_type_dict = None, raise_errors = True, **kwargs):
        """
        Initialize the rowIndex transformer.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """

        self.value_maker = self.ValueMaker(raise_errors=raise_errors)

        super().__init__(properties_of, self.value_maker, label_maker, branching_properties, columns, output_validator,
                         multi_type_dict, raise_errors=raise_errors, **kwargs)

    def __call__(self, row, i, result_object=None):
        """
        Process a row and yield the row index as node ID.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Returns:
            int: The row index if valid.

        Raises:
            Warning: If the row index is invalid.
        """
        for value in self.value_maker(self.columns, row, i):
            yield self.create(value, row)

class map(base.Transformer):
    """Transformer subclass used for the simple mapping of cell values of defined columns and creating
    nodes with their respective values as id."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            for key in columns:
                if key not in row:
                    self.error(f"Column '{key}' not found in data", section="map.call",
                               exception=exceptions.TransformerDataError)
                yield row[key]

    def __init__(self, properties_of, label_maker = None, branching_properties = None, columns=None, output_validator: validate.OutputValidator = None, multi_type_dict = None, raise_errors = True, **kwargs):
        """
        Initialize the map transformer.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """

        self.value_maker = self.ValueMaker(raise_errors=raise_errors)

        super().__init__(properties_of, self.value_maker, label_maker, branching_properties, columns, output_validator,
                         multi_type_dict, raise_errors=raise_errors, **kwargs)

    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The cell value if valid.

        Raises:
            Warning: If the cell value is invalid.
        """
        if not self.columns:
            self.error(f"No column declared for the {type(self).__name__} transformer, did you forgot to add a `columns` keyword?", section="map.call", exception = exceptions.TransformerInputError)

        for value in self.value_maker(self.columns, row, i):
            yield self.create(value, row)


class translate(base.Transformer):
    """Translate the targeted cell value using a tabular mapping and yield a node with using the translated ID."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            for key in columns:
                if key not in row:
                    self.error(f"Column '{key}' not found in data", section="map.call",
                               exception=exceptions.TransformerDataError)
                yield row[key]

    def __init__(self, properties_of, label_maker = None, branching_properties = None, columns=None, output_validator: validate.OutputValidator = None, multi_type_dict = None, raise_errors = True, **kwargs):
        """
        Constructor.

        NOTE: The user should provide at least either `translations` or `translations_file`, but not both.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            translations: A dictionary figuring what to replace (keys) with which string (values).
            translations_file: A filename pointing to a tabular file readable by Pandas' csv_read.
            translate_from: The column in the file containing what to replace.
            translate_to: The column in the file containing the replacement string.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
            kwargs: Additional arguments to pass to Pandas' read_csv (if "sep=TAB", reads the translations_file as tab-separated).
        """

        self.value_maker = self.ValueMaker(raise_errors=raise_errors)

        super().__init__(properties_of, self.value_maker, label_maker, branching_properties, columns, output_validator,
                         multi_type_dict, raise_errors=raise_errors, **kwargs)
        self.map = map(properties_of, label_maker, branching_properties, columns, output_validator, multi_type_dict, **kwargs)

        # Since we cannot expand kwargs, let's recover what we have inside.
        translations = kwargs.get("translations", None)
        translations_file = kwargs.get("translations_file", None)
        translate_from = kwargs.get("translate_from", None)
        translate_to = kwargs.get("translate_to", None)

        if translations and translations_file:
            self.error(f"Cannot have both `translations` (=`{translations}`) and `translations_file` (=`{translations_file}`) defined in a {type(self).__name__} transformer.", secton="translate", exception = exceptions.TransformerInterfaceError)

        if translations:
            self.translate = translations
            logger.debug(f"\t\t\tManual translations: `{self.translate}`")
        elif translations_file:
            logger.debug(f"\t\t\tGet translations from file: `{translations_file}`")
            if not translate_from:
                self.error(f"No translation source column declared for the `{type(self).__name__}` transformer using translations_file=`{translations_file}`, did you forget to add a `translate_from` keyword?", section="translate.init", exception = exceptions.TransformerInterfaceError)
            if not translate_to:
                self.error(f"No translation target column declared for the `{type(self).__name__}` transformer using translations_file=`{translations_file}`, did you forget to add a `translate_to` keyword?", section="translate.init", exception = exceptions.TransformerInterfaceError)
            else:
                self.translations_file = translations_file
                self.translate_from = translate_from
                self.translate_to = translate_to

                # Extract available arguments from Pandas' read_csv docstring:
                pd_read_csv_args = []
                for line in pd.read_csv.__doc__.split("\n"):
                    if re.match(r"^[a-z_]+ :", line):
                        pd_read_csv_args.append(line.split(":")[0].strip())

                # Keep only the user-passed arguments that are in Pandas' read_csv list.
                pd_args = {k:v for k,v in kwargs.items() if k in pd_read_csv_args}

                if "sep" in pd_args and pd_args["sep"] == "TAB":
                    logger.debug(f"\t\t\tMapping asked for sep:TAB, enable Pandas' read_csv engine:python to avoid a warning.")
                    pd_args["sep"] = '\t'
                    pd_args["engine"] = "python"

                logger.debug(f"\t\t\tArguments passed to pandas.read_csv: `{pd_args}`")

                self.df = pd.read_csv(self.translations_file, **pd_args)

                if self.translate_from not in self.df.columns:
                    self.error(f"Source column `{self.translate_from}` not found in {type(self).__name__} transformer’s translations file `{self.translations_file}`, available headers: `{','.join(self.df.columns)}`.", section="translate.init", exception = exceptions.TransformerDataError)

                if self.translate_to not in self.df.columns:
                    self.error(f"Target column `{self.translate_to}` not found in {type(self).__name__} transformer’s translations file `{self.translations_file}`, available headers: `{','.join(self.df.columns)}`.", section="translate.init", exception = exceptions.TransformerDataError)

                self.translate = {}
                for i,row in self.df.iterrows():
                    if row[self.translate_from] and row[self.translate_to]:
                        self.translate[row[self.translate_from]] = row[self.translate_to]
                    else:
                        logger.warning(f"Cannot translate from `{self.translate_from}` to `{self.translate_to}`, invalid translations values at row {i} of file `{self.translations_file}`: `{row[self.translate_from]}` => `{row[self.translate_to]}`. I will ignore this translation.")

        else:
            self.error(f"When using a {type(self).__name__} transformer, you must define either `translations` or `translations_file`.", section="translate.init", exception = exceptions.TransformerInterfaceError)


        if not self.translate:
            self.error(f"No translation found, did you forget the `translations` keyword?", section="translate.init", exception = exceptions.TransformerInterfaceError)

    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The cell value if valid.

        Raises:
            Warning: If the cell value or the translation is invalid.
        """
        if not self.columns:
            self.error(f"No column declared for the {type(self).__name__} transformer, did you forgot to add a `columns` keyword?", section="translate", exception = exceptions.TransformerDataError)

        for key in self.columns:
            if key not in row:
                self.error(f"Column '{key}' not found in data", section="translate", exception = exceptions.TransformerDataError)
            cell = row[key]
            if cell in self.translate:
                row[key] = self.translate[cell]
            else:
                logger.warning(f"Row {i} does not contain something to be translated from `{self.translate_from}` to `{self.translate_to}` at column `{key}`.")

        for e, edge, node in self.map(row, i):
            yield e, edge, node

class string(base.Transformer):
    """A transformer that makes up the given static string instead of extractsing something from the table."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True, string: str = None):
            self.string = string
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            yield self.string

    def __init__(self, properties_of, label_maker = None, branching_properties = None, columns=None, output_validator: validate.OutputValidator = None, multi_type_dict = None, raise_errors = True, **kwargs):
        """
        Constructor.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            value: The string to use.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """

        self.value = kwargs.get("value", None)
        self.value_maker = self.ValueMaker(raise_errors=raise_errors, string=self.value)

        super().__init__(properties_of, self.value_maker, label_maker, branching_properties, columns, output_validator,
                         multi_type_dict, raise_errors=raise_errors, **kwargs)

    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The cell value if valid.

        Raises:
            Warning: If the cell value is invalid.
        """
        if not self.value:
            self.error(f"No value passed to the {type(self).__name__} transformer, did you forgot to add a `value` keyword?", section="string.call", exception = exceptions.TransformerInterfaceError)

        for value in self.value_maker(self.columns, row, i):
            yield self.create(value, row)

class replace(base.Transformer):
    """Transformer subclass used to remove characters that are not allowed from cell values of defined columns.
     The forbidden characters are defined by a regular expression pattern, and are substituted with a user-defined
     character or removed entirely. In case the cell value is made up of only forbidden characters, the node is not
     created and a warning is logged."""

    class ValueMaker(make_value.ValueMaker):

        def __init__(self, raise_errors: bool = True, forbidden = None, substitute = None):

            self.forbidden = forbidden
            self.substitute = substitute
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            for key in columns:
                logger.info(
                    f"Setting forbidden characters: {self.forbidden} for `replace` transformer, with substitute character: `{self.substitute}`.")
                formatted = re.sub(self.forbidden, self.substitute, row[key])
                strip_formatted = formatted.strip(self.substitute)
                logger.debug(f"Formatted value: {strip_formatted}")
                yield strip_formatted

    def __init__(self, properties_of, label_maker = None, branching_properties = None, columns=None, output_validator: validate.OutputValidator = None, multi_type_dict = None, raise_errors = True, **kwargs):
        """
        Constructor.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            forbidden: The regular expression pattern to match forbidden characters.
            substitute: The string to replace forbidden characters with.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """
        self.forbidden = kwargs.get("forbidden", r'[^a-zA-Z0-9_`.()]') # By default, allow alphanumeric characters (A-Z, a-z, 0-9),
        # underscore (_), backtick (`), dot (.), and parentheses (). TODO: Add or remove rules as needed based on errors in Neo4j import.
        self.substitute = kwargs.get("substitute", "")

        self.value_maker = self.ValueMaker(raise_errors=raise_errors, forbidden=self.forbidden, substitute=self.substitute)

        super().__init__(properties_of, self.value_maker, label_maker, branching_properties, columns, output_validator,
                         multi_type_dict, raise_errors=raise_errors, **kwargs)

    def __call__(self, row, i):
        """
        Process a row and yield cell values with forbidden characters removed or replaced.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The processed cell value with forbidden characters removed or replaced.

        Raises:
            Warning: If the processed cell value is invalid.
        """
        for value in self.value_maker(self.columns, row, i):
            yield self.create(value, row)


import logging
import re
from collections.abc import Iterable, Generator
from abc import ABCMeta as ABSTRACT, ABCMeta, abstractmethod
from abc import abstractmethod as abstract
from typing import TypeAlias
from typing import Optional
import pandas as pd
import pandera as pa
from jinja2.compiler import has_safe_repr

from . import errormanager
from . import validate
from . import serialize
from . import exceptions
from . import make_value
from .validate import SkipValidator

logger = logging.getLogger("ontoweaver")

# TODO? Strategy using a user defined __eq__ method, enabling a more flexible comparison of objects, but in O(n2).
# class Comparer(metaclass=ABCMeta):
#     @abstractmethod
#     def __call__(self, elem1, elem2):
#         raise NotImplementedError
#
# class CompEq(Comparer):
#     def __call__(self, elem1, elem2):
#         return elem1 is elem2
# FIXME use hash functions for comparison.

class ErrorManager:
    def __init__(self, raise_errors = True):
        self.raise_errors = raise_errors

    def error(self, msg, section = None, index = None, exception = RuntimeError, indent = 0):
        location = ""
        if section:
            location = f" [for {section}"
            if index:
                location += f" #{index}"
            location += "]"

        err = "\t"*indent
        err += msg
        err += location

        logger.error(err)

        if self.raise_errors:
            raise exception(err)

        return err


class Element(metaclass = ABSTRACT):
    """Base class for either Node or Edge.

    Manages allowed properties mechanics."""

    def __init__(self,
                 id        : Optional[str] = None,
                 properties: Optional[dict[str,str]] = {},
                 label     : Optional[str] = None,
                 serializer: Optional[serialize.Serializer] = serialize.All(),
                 ):
        """Instantiate an element.

        :param str id: Unique identifier of the element. If id == None, is then set to the empty string.
        :param dict[str,str] properties: All available properties for this instance.
        :param str label: The label of the element.
        :param Comparer serializer: The comparer to use for equality checks. Default uses the python `is` operator.
        """
        if not id:
            self._id = ''
        else:
            self._id = str(id)

        # Use the setter to get sanity checks.
        self.properties = properties

        if not label:
            # Do not change the name here, or BioCypher will have problem
            # finding back labels.
            self._label = self.__class__.__name__ 
        else:
            self._label = str(label)

        self.serializer = serializer

    def __str__(self):
        return self.serializer(self)

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    @staticmethod
    @abstract
    def fields() -> list[str]:
        """List of property fields provided by the (sub)class."""
        raise NotImplementedError

    @abstract
    def as_tuple(self):
        """Convert the element class into Biocypher's expected tuple.

        Filter out properties along the way.
        """
        raise NotImplementedError

    @classmethod
    @abstract
    def from_tuple(cls,
                   biocypher_tuple : tuple,
                   serializer: Optional[serialize.Serializer] = serialize.All()
        ):
        # return cls(biocypher_tuple,serializer)
        raise NotImplementedError

    @property
    def id(self) -> str:
        return self._id

    @property
    def label(self) -> str:
        return self._label

    @property
    def properties(self) -> dict[str,str]:
        return self._properties

    @properties.setter
    def properties(self, properties: dict[str,str]):
        """Set available properties.

        Asserts that the passed properties are in the declared fields."""

        # Sanity checks:
        assert(properties is not None)
        assert(type(properties) == dict)
        # logger.debug(f"Properties of `{type(self).__name__}`: {list(properties.keys())}, available: {list(self.available())}")
        # TODO enable the usage of available() function to disable / enable parts of ontology / certain nodes
        # for p in properties:
        #     if p not in self.available():
        #         logger.error(f"\t\tProperty `{p}` should be available for type `{type(self).__name__}`, available ones: `{list(self.available())}`")
        #         assert(p in self.available())
        self._properties = properties


class Node(Element):
    """Base class for any Node."""

    def __init__(self,
                 id        : Optional[str] = None,
                 properties: Optional[dict[str,str]] = {},
                 label     : Optional[str] = None,  # Set from subclass name.
                 serializer: Optional[serialize.Serializer] = serialize.node.All(),
                 ):
        """Instantiate a Node.

        :param str id: Unique identifier of the node. If id == None, is then set to the empty string.
        :param dict[str,str] properties: All available properties for this instance.
        :param str label: The label of the node.
        :param Comparer serializer: The comparer to use for equality checks. Default uses the python `is` operator.
        """
        super().__init__(id = id, properties = properties, label = label, serializer = serializer)

    Tuple: TypeAlias = tuple[str,str,dict[str,str]]
    def as_tuple(self) -> Tuple:
        """Export the Node as a Biocypher tuple."""
        return (
            self._id,
            self._label,
            # FIXME this has been changed to keep ALL properties. No checking if allowed
            self.properties
        )

    @classmethod
    def from_tuple(cls,
                   biocypher_tuple : tuple[str,str,dict[str,str]],
                   serializer: Optional[serialize.Serializer] = serialize.node.All(),
                   ):
        assert(len(biocypher_tuple) == 3)
        return cls(
            id         = biocypher_tuple[0],
            label      = biocypher_tuple[1],
            properties = biocypher_tuple[2],
            serializer = serializer)

    def fields(self) -> list[str]:
        """List of property fields provided by the (sub)class."""
        return list(self.properties.keys())

    def __str__(self):
        return self.serializer(self)

    def __repr__(self):
        return f"<['{self.label}':'{self.id}'/{self.properties}]>"

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    # def serialize(self):
    #     return {
    #         "id": self._id,
    #         "label": self._label,
    #         "properties": self.properties,
    #         "serializer": self.serializer
    #     }

class Edge(Element):
    """Base class for any Edge."""

    def __init__(self,
                 id        : Optional[str] = None,
                 id_source : Optional[str] = None,
                 id_target : Optional[str] = None,
                 properties: Optional[dict[str,str]] = {},
                 label     : Optional[str] = None,  # Set from subclass name.
                 serializer: Optional[serialize.Serializer] = serialize.edge.All(),
                 ):
        """Instantiate an Edge.

        :param str id: Unique identifier of the edge. If id == None, is then set to the empty string.
        :param str id_source: Unique identifier of the source Node. If None, is then set to the empty string.
        :param str id_target: Unique identifier of the target Node. If None, is then set to the empty string.
        :param dict[str,str] properties: All available properties for this instance.
        :param str label: The label of the node.
        :param Comparer comparer: The comparer to use for equality checks. Default uses the python `is` operator.
        """
        super().__init__(id = id, properties = properties, label = label, serializer = serializer)
        self._id_source = str(id_source)
        self._id_target = str(id_target)

    @staticmethod
    @abstract
    def source_type():
        raise NotImplementedError

    @staticmethod
    @abstract
    def target_type():
        raise NotImplementedError

    @property
    def id_source(self):
        return self._id_source

    @id_source.setter
    def id_source(self, id_source):
        self._id_source = id_source

    @property
    def id_target(self):
        return self._id_target

    @id_target.setter
    def id_target(self, id_target):
        self._id_target = id_target

    Tuple: TypeAlias = tuple[str,str,str,dict[str,str]]
    def as_tuple(self) -> Tuple:
        """Export the Edge as a Biocypher tuple."""
        return (
            self._id,
            self._id_source,
            self._id_target,
            self._label,
            #.FIXME no checking if properties are allowed allowed_properties()
            self.properties
        )

    @classmethod
    def from_tuple(cls,
                   biocypher_tuple : tuple[str,str,str,str,dict[str,str]],
                   serializer: Optional[serialize.Serializer] = serialize.edge.All()
                   ):
        assert(len(biocypher_tuple) == 5)
        return cls(
            id         = biocypher_tuple[0],
            id_source  = biocypher_tuple[1],
            id_target  = biocypher_tuple[2],
            properties = biocypher_tuple[4],
            label      = biocypher_tuple[3],
            serializer = serializer)

    def __repr__(self):
        if self.source_type() == Node:
            st = "."
        else:
            st = f"’{self.source_type()}’"
        if self.target_type() == Node:
            tt = "."
        else:
            tt = f"’{self.target_type()}’"

        return f"<[{st}:'{self.id_source}']--('{self.label}':'{self.id}'/{self.properties})-->[{tt}:'{self.id_target}']>"

    def fields(self) -> list[str]:
        """List of property fields provided by the (sub)class."""
        return list(self.properties.keys())

    def __str__(self):
        return self.serializer(self)

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    # def serialize(self):
    #     return {
    #         "id": self._id,
    #         "id_source": self._id_source,
    #         "id_target": self._id_target,
    #         "label": self._label,
    #         "properties": self.properties,
    #         "serializer": self.serializer
    #     }

class GenericEdge(Edge):
    """Base class for any Edge."""

    def __init__(self,
                 id        : Optional[str] = None,
                 id_source : Optional[str] = None,
                 id_target : Optional[str] = None,
                 properties: Optional[dict[str,str]] = {},
                 label     : Optional[str] = None,  # Set from subclass name.
                 serializer: Optional[serialize.Serializer] = serialize.edge.All(),
                 ):
        """Instantiate an Edge.

        :param str id: Unique identifier of the edge. If id == None, is then set to the empty string.
        :param str id_source: Unique identifier of the source Node. If None, is then set to the empty string.
        :param str id_target: Unique identifier of the target Node. If None, is then set to the empty string.
        :param dict[str,str] properties: All available properties for this instance.
        :param str label: The label of the node.
        :param Comparer comparer: The comparer to use for equality checks. Default uses the python `is` operator.
        """
        super().__init__(id = id, id_source = id_source, id_target = id_target, properties = properties, label = label, serializer = serializer)

    @staticmethod
    def source_type():
        return Node

    @staticmethod
    def target_type():
        return Node


class Adapter(errormanager.ErrorManager, metaclass = ABSTRACT):
    """Base class for implementing a canonical Biocypher adapter."""

    def __init__(self, raise_errors = True
    ):
        """Allow to indicate which Element subclasses and which property fields
        are allowed to be exported by Biocypher.

        :param Iterable[Node] node_types: Allowed Node subclasses.
        :param Iterable[Edge] edge_types: Allowed Edge subclasses.
        """

        self._nodes = []
        self._edges = []
        self.errors = []
        super().__init__(raise_errors)

    def nodes_append(self, node_s) -> None:
        """Append an Node (or each Node in a list of nodes) to the internal list of nodes."""
        if issubclass(type(node_s), Node):
            nodes = [node_s]
        else:
            nodes = node_s

        # logger.debug(f"Nodes: {nodes}.")
        for node in nodes:
            # logger.debug(f"\tAppend node {node}.")
            # Checking for duplicates in reconciliation, otherwise complexity too high.
            self._nodes.append(node.as_tuple())
            # return True

    def edges_append(self, edge_s) -> None:
        """Append an Edge (or each Edge in a list of edges) to the internal list of edges."""
        if issubclass(type(edge_s), Edge):
            edges = [edge_s]
        else:
            edges = edge_s

        # logger.debug(f"Edges: {edges}.")
        for edge in edges:
            # logger.debug(f"\tAppend edge {edge}.")
            # Checking for duplicates in reconciliation, otherwise complexity too high.
            self._edges.append(edge.as_tuple())
            # return True

    @property
    def nodes(self) -> Iterable[Node.Tuple]:
        """Return a generator yielding nodes."""
        for n in self._nodes:
            yield n

    @property
    def edges(self) -> Iterable[Edge.Tuple]:
        """Return a generator yielding edges."""
        for e in self._edges:
            yield e


class Transformer(errormanager.ErrorManager):
    """"Class used to manipulate cell values and return them in the correct format."""""

    def __init__(self, properties_of, value_maker = None, label_maker = None, branching_properties = None, columns = None,
                 output_validator: validate.OutputValidator() = None, multi_type_dict = None, raise_errors = True, **kwargs):
        """
        Instantiate transformers.

        :param properties_of: the properties of each node type.
        :param value_maker: the ValueMaker object used for the logic of cell value selection for each transformer. Default is None.
        :param label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
        :param branching_properties: in case of branching on cell values, the dictionary holds the properties for each branch.
        :param columns: the columns to use in the mapping.
        :param output_validator: the OutputValidator object used for validating transformer output. Default is None.
        :param multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
        :param raise_errors: whether to raise errors or not. Default is True.
        each transformer is instantiated with a default OutputValidator object, and additional user-defined rules if needed in
        the tabular module.

        """
        super().__init__(raise_errors)

        self.properties_of = properties_of
        self.value_maker = value_maker
        self.label_maker = label_maker
        self.branching_properties = branching_properties
        self.columns = columns
        self.output_validator = output_validator
        if not self.output_validator:
            self.output_validator = validate.OutputValidator(validate.default_validation_rules, raise_errors = raise_errors)
        self.parameters = kwargs
        self.multi_type_dict = multi_type_dict
        self.final_type = None # The final type is to be passed by the label maker class based on the YAML mapping. That
                               # is why here it is None by default
        self.kwargs = kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_transformer(self):
        return self

    @abstract
    def __call__(self, row, i):
        raise NotImplementedError

    #FIXME: The functions below are never implemented.
    @abstract
    def nodes(self):
        raise NotImplementedError

    @abstract
    def edges(self):
        raise NotImplementedError

    @staticmethod
    @abstract
    def edge_type():
        raise NotImplementedError

    @staticmethod
    @abstract
    def target_type():
       raise NotImplementedError

    @classmethod
    def source_type(cls):
       return cls.edge_type().source_type()

    def __repr__(self):

        representation = ""

        if hasattr(self, "from_subject"):
            from_subject = self.from_subject
        else:
            from_subject = "."

        target_name = ""
        edge_name = ""

        if hasattr(self, "multi_type_dict") and self.multi_type_dict is not None:
            for key, value in self.multi_type_dict.items():
                if value['to_object'] and value['via_relation']:
                    target_name = value['to_object']
                    edge_name = value['via_relation']
                elif value['to_object'] and not value['via_relation']:
                    target_name = value['to_object']
                    edge_name = "."


                if self.properties_of:
                    # The transformer is not a branching transformer, and has only one set of properties.
                    props = self.properties_of
                elif self.branching_properties and self.branching_properties.get(value['to_object'], None):
                    # The transformer is a branching transformer, and has multiple sets of properties. We extract the ones for the current type.
                    props = self.branching_properties.get(value['to_object'])
                else:
                    # The transformer has no properties for the type.
                    props = "{}"


                params = ""
                parameters = {k:v for k,v in self.parameters.items() if k not in ['subclass', 'from_subject', "match"]}
                if parameters:
                    p = []
                    for k,v in parameters.items():
                        p.append(f"{k}={v}")
                    params = ','.join(p)

                if from_subject == "." and edge_name == "." and target_name == "." and props == "{}":
                    # If this is a property transformer
                    link = ""

                elif from_subject == "." and edge_name == "." and (target_name != "." or props != "{}"):
                    # This a subject transformer.
                    link = f" => [{target_name}/{props}]"

                else:
                    # This is a regular transformer.
                    link = f" => [{from_subject}]--({edge_name})->[{target_name}/{props}]"

                if self.columns:
                    columns = self.columns
                else:
                    columns = []

                for c in columns:
                    if type(c) != str:
                        self.error(f"Column `{c}` is not a string, did you mistype a leading colon?", exception=exceptions.ParsingError)

                representation += (f"<Transformer:{type(self).__name__}({params}) {','.join(columns)}{link}>")

        else:
            #The transformer is a property transformer. We add the property name and value in the tabular.properties() function.
            representation += (f"<Transformer:{type(self).__name__}() {','.join(self.columns) if self.columns else ''} =>")

        return representation

    def validate(self, res):
        """
        Validate the output of the transformer, using the output_validator. of the transformer instance.
        """
        try:
            if isinstance(self.output_validator, validate.SimpleOutputValidator) or isinstance(self.output_validator, SkipValidator):
                # The SimpleOutputValidator and SkipValidator do not use the Pandera package, and the value is therefore not
                # required to be in a DataFrame format. Voiding the creation of a DataFrame here saves computational time for
                # large datasets.
                if self.output_validator(res):
                    return True
                else:
                    return False
            else:
                if self.output_validator(pd.DataFrame([res], columns=["cell_value"])):
                    return True
                else:
                    return False
        except pa.errors.SchemaErrors as error:
            msg = f"Transformer {self.__repr__()} did not produce valid data {error}."
            self.error(msg, exception = exceptions.DataValidationError)

    def create(self, returned_value, row):
        """
        Create the output of the transformer, using the label_maker of the transformer instance.

        Returns:
            Extracted cell value (can be node ID, property value, edge ID), edge type and target node type.
        """
        result_object = self.label_maker(self.validate, returned_value, self.multi_type_dict, self.branching_properties, row)
        if result_object.target_node_type:
            self.target_type = result_object.target_node_type.__name__
        if result_object.target_element_properties is not None:
            self.properties_of = result_object.target_element_properties
        self.final_type = result_object.final_type
        return result_object.extracted_cell_value, result_object.edge_type, result_object.target_node_type


class All:
    """Gathers lists of subclasses of Element and their fields
    existing in a given module.

    Is generally used to label_maker an `all` variable in a module:
    .. code-block:: python

        all = base.All(sys.modules[__name__])

    Which can later be used to pass all available Element types to an Adapter:
    .. code-block:: python

        a = MyAdapter( node_types = MyModule.all.nodes() )
    """

    def __init__(self, module):
        self.module = module

    def elements(self, asked: Element = Element) -> list[Element]:
        m = self.module.__dict__
        classes = []
        for c in m:
            if isinstance(m[c], type) \
            and m[c].__module__ == self.module.__name__ \
            and issubclass(m[c], asked):
                classes.append(m[c])
                logger.debug(f"Found `{asked.__name__}` class: `{m[c]}` (prop: `{m[c].fields()}`).")
                # t = m[c]
                # logger.debug(f"\t\t#### {t.mro()[:-3]}/{t.__name__} => {t.fields()}")
        return classes

    def nodes(self) -> list[Node]:
        return self.elements(Node)

    def edges(self) -> list[Edge]:
        return self.elements(Edge)

    def node_fields(self) -> list[str]:
        names = [] # FIXME use a set?
        for c in self.nodes():
            names += c.fields()
        return names

    def edge_fields(self) -> list[str]:
        names = [] # FIXME use a set?
        for c in self.edges():
            names += c.fields()
        return names
